"""
Test search classes
"""

import numpy as np

from scipy.stats.distributions import expon
from sklearn.linear_model import LogisticRegression
from sklearn.naive_bayes import GaussianNB
from sklearn.ensemble import RandomForestClassifier

try:
    from skdist.distribute import search
    from skdist.distribute.search import (
        DistGridSearchCV, DistRandomizedSearchCV,
        DistMultiModelSearch
        )
    _import_error = None
except Exception as e:
    _import_error = e

def test_search():
    assert _import_error == None

def test_gs():
    X = np.array([[1,1,1], [0,0,0], [-1,-1,-1]]*100)
    y = np.array([0,0,1]*100)
    gs = DistGridSearchCV(
        LogisticRegression(solver="liblinear"), 
        {"C": [0.1, 1.0]}, cv=3
        )
    gs.fit(X,y)
    preds = gs.predict(X[:3])
    assert np.allclose(preds, np.array([0,0,1]))

def test_rs():
    X = np.array([[1,1,1], [0,0,0], [-1,-1,-1]]*100)
    y = np.array([0,0,1]*100)
    rs = DistRandomizedSearchCV(
        LogisticRegression(solver="liblinear"),
        {"C": [0.1, 1.0]}, cv=3, n_iter=2
        )
    rs.fit(X,y)
    preds = rs.predict(X[:3])
    assert np.allclose(preds, np.array([0,0,1]))

def test_multimodel():
    X = np.array([[1,1,1], [0,0,0], [-1,-1,-1]]*100)
    y = np.array([0,0,1]*100)
    models = [
        ("rf", RandomForestClassifier(n_estimators=10), {"max_depth": [5,10], "min_samples_split": range(2, 10)}),
        ("lr0", LogisticRegression(multi_class="auto", solver="liblinear"), {"C": [0.1, 1.0, 10]}), 
        ("lr1", LogisticRegression(multi_class="auto", solver="liblinear"), {"fit_intercept": [True, False], "C": expon()}), 
        ("nb", GaussianNB(), {})
        ]
    clf = DistMultiModelSearch(models, n=2)
    clf.fit(X,y)
    preds = clf.predict(X[:3])
    assert np.allclose(preds, np.array([0,0,1]))



"""
Test feature eliminator classes
"""

import numpy as np

from scipy.sparse import csr_matrix
from sklearn.linear_model import LogisticRegression
from sklearn import datasets

try:
    from skdist.distribute import eliminate
    from skdist.distribute.eliminate import DistFeatureEliminator 
    _import_error = None
except Exception as e:
    _import_error = e

def test_import():
    assert _import_error == None

def test_fit():
    X,y = datasets.load_iris(return_X_y=True)
    fe = DistFeatureEliminator(
        LogisticRegression(solver="liblinear", multi_class="auto"), 
        scoring="f1_weighted", min_features_to_select=3, cv=3
        )
    fe.fit(X,y)
    assert np.allclose(fe.best_features_, [1,2,3])

def test_score():
    X,y = datasets.load_iris(return_X_y=True)
    fe = DistFeatureEliminator(
        LogisticRegression(solver="liblinear", multi_class="auto"),
        scoring="f1_weighted", min_features_to_select=2, cv=3
        )
    fe.fit(X,y)
    assert round(fe.score(X,y),2) == 0.95

def test_predict():
    X,y = datasets.load_iris(return_X_y=True)
    fe = DistFeatureEliminator(
        LogisticRegression(solver="liblinear", multi_class="auto"),
        scoring="f1_weighted", min_features_to_select=1, cv=3, step=2
        )
    fe.fit(X,y)
    assert fe.predict(X)[0] == 0

def test_sparse():
    X,y = datasets.load_iris(return_X_y=True)
    fe = DistFeatureEliminator(
        LogisticRegression(solver="liblinear", multi_class="auto"),
        scoring="f1_weighted", min_features_to_select=3, cv=3
        )
    fe.fit(csr_matrix(X),y)
    assert fe.predict(X)[0] == 0



"""
Test multiclass classes
"""

try:
    import numpy as np
    from sklearn.linear_model import LogisticRegression
    from skdist.distribute import multiclass
    from skdist.distribute.multiclass import (
        DistOneVsRestClassifier, DistOneVsOneClassifier  
        )
    _import_error = None
except Exception as e:
    _import_error = e

def test_multiclass():
    assert _import_error == None

def test_ovr():
    X = np.array([[0,0,1,1], [1,1,0,0], [-1,-1,-1,-1]]*100)
    y = np.array([0,1,2]*100)
    ovr = DistOneVsRestClassifier(LogisticRegression(solver="liblinear"))
    ovr.fit(X,y)
    preds = ovr.predict(X[:3])
    assert np.allclose(preds, np.array([0,1,2]))

def test_ovo():
    X = np.array([[0,0,1,1], [1,1,0,0], [-1,-1,-1,-1]]*100)
    y = np.array([0,1,2]*100)
    ovo = DistOneVsOneClassifier(LogisticRegression(solver="liblinear"))
    ovo.fit(X,y)
    preds = ovo.predict(X[:3])
    assert np.allclose(preds, np.array([0,1,2]))



"""
Test base imports
"""

try:
    from skdist.distribute import base
    from skdist.distribute.base import *
    from skdist.distribute.ensemble import DistRandomForestClassifier
    _import_error = None
except Exception as e:
    _import_error = e

def test_import_base():
    assert _import_error == None

def test_clone():
    rf = DistRandomForestClassifier(n_estimators=10)
    rf_cloned = base._clone(rf)
    assert rf.n_estimators == rf_cloned.n_estimators



"""
Test top level import
"""

try:
    from skdist import *
    _top_import_error = None
except Exception as e:
    _top_import_error = e

def test_import_skdist():
    assert _top_import_error == None



"""
Distributed scikit-learn meta-estimators in PySpark
===================================================
skdist is a Python module that aims to efficiently distribute scikit-learn
meta-estimator model fitting with PySpark. Where sklearn uses joblib to
parallelize these operations, distml uses spark, enabling effectively infinite
parallelization, unconstrained by the number of cores on a single machine.

Existing scikit-learn meta-estimators for model selection, ensembling,
multiclass fitting, and feature unioning are inherited, and have their fit methods
overridden with a distributed implementation using spark. The resulting fitted model
is effectively identical to its single machine counterpart,
ready for prediction using the inherited methods.
"""

__version__ = '0.1.4'

__all__ = ['distribute', 'preprocessing', 'postprocessing', 'tests']



"""
Distributed multiclass meta-estimators
"""

import numpy as np
import pandas as pd
import warnings

from joblib import Parallel, delayed
from copy import copy
from collections.abc import Sequence
from scipy.sparse import vstack
from sklearn.base import BaseEstimator
from sklearn.multiclass import OneVsRestClassifier, OneVsOneClassifier
from sklearn.preprocessing import LabelBinarizer, MultiLabelBinarizer
from sklearn.preprocessing import normalize
from sklearn.utils import shuffle
from sklearn.utils.validation import check_X_y
from sklearn.utils.multiclass import check_classification_targets, type_of_target
from sklearn.model_selection import train_test_split, GroupKFold

from .validation import _check_estimator
from .utils import _safe_split
from .base import (
    _clone, _get_value, _parse_partitions
    )

__all__ = [
    "DistOneVsRestClassifier",
    "DistOneVsOneClassifier"
]

def _chunks(l, n):
    """ Yield successive n-sized chunks from input list """
    for i in range(0, len(l), n):
        yield l[i:i + n]

def _split_X(X, n_splits, sc):
    """ Split and broadcast a sparse matrix """
    if n_splits > 1:
        if isinstance(X, pd.DataFrame):
            inds = list(range(len(X)))
            inds_list = list(_chunks(inds, int(len(inds) / float(n_splits))))
            Xs = [X.ix[slc] for slc in inds_list]
        else:
            # X is a scipy sparse matrix
            inds = list(range(X.shape[0]))
            inds_list = list(_chunks(inds, int(len(inds) / float(n_splits))))
            Xs = [X[slc, :] for slc in inds_list]

        return [sc.broadcast(x) for x in Xs]
    else:
        return X

def _combine_Xs(Xs):
    """ Combine and access values of broadcasted, splitted sparse matrix """
    if isinstance(Xs, list):
        if isinstance(Xs[0].value, pd.DataFrame):
            df = pd.DataFrame()
            return df.append([x.value for x in Xs])
        else:
            return vstack([x.value for x in Xs])
    else:
        return Xs

def _use_best_estimator(x):
    """ Pull best estimator out of meta-estimator if available """
    estimator_ = x.best_estimator_ if hasattr(x, "best_estimator_") else x
    if hasattr(x, "cv_results_"):
        estimator_.cv_results_ = pd.DataFrame(x.cv_results_)
        for col in estimator_.cv_results_.columns:
            estimator_.cv_results_[col] = estimator_.cv_results_[col].astype(str)
        estimator_.cv_results_ = estimator_.cv_results_.to_dict("list")
    return estimator_

def _negatives_mask(X, y, max_negatives=None, random_state=None, method="ratio"):
    """ Limit the number of negative records in training set """
    if max_negatives is None:
        return [X, y]
    else:
        pos_mask = y == 1
        if method == "ratio":
            pass
        elif method == "multiplier":
            max_negatives = int(max_negatives * len(y[pos_mask]))
        else:
            raise ValueError("Unknown method. Options are 'ratio' or 'multiplier'.")
        if isinstance(max_negatives, int) and max_negatives >= len(y[~pos_mask]):
            return [X, y]
        max_negatives = (
            max_negatives
            if isinstance(max_negatives, float)
            else (max_negatives / float(len(y[~pos_mask]))))
        _, X_neg, _, y_neg = train_test_split(
            X[~pos_mask, :], y[~pos_mask],
            test_size=max_negatives, random_state=random_state)
        stack_func = vstack if hasattr(X, "nnz") else np.vstack
        return shuffle(
            stack_func([X[pos_mask, :], X_neg]), np.concatenate([y[pos_mask], y_neg]),
            random_state=random_state)

def _fit_binary(estimator, X, y, fit_params, classes=None, max_negatives=None, random_state=None, method="ratio"):
    """ Fit a single binary estimator """
    X = _combine_Xs(X)
    unique_y = np.unique(y)
    if len(unique_y) == 1:
        if classes is not None:
            if y[0] == -1:
                c = 0
            else:
                c = y[0]
            warnings.warn("Label %s is present in all training examples." %
                          str(classes[c]))
        est = _ConstantPredictor().fit(*_negatives_mask(
            (X.value if hasattr(X, "value") else X), unique_y,
            max_negatives=max_negatives, random_state=random_state,
            method=method))
    else:
        est = _clone(_get_value(estimator))
        est.fit(*_negatives_mask(
            (X.value if hasattr(X, "value") else X), y,
            max_negatives=max_negatives, random_state=random_state,
            method=method), **fit_params)
    return _use_best_estimator(est), classes[1]

def _fit_ovo_binary(estimator, X, y, i, j, fit_params):
    """ Fit a single binary estimator (one-vs-one) """
    cond = np.logical_or(y == i, y == j)
    y_ = y[cond]
    y_binary = np.empty(y_.shape, np.int)
    y_binary[y_ == i] = 0
    y_binary[y_ == j] = 1
    indcond = np.arange(X.shape[0])[cond]
    return _fit_binary(estimator,
                       _safe_split(estimator, X, None, indices=indcond)[0],
                       y_binary, fit_params, classes=[i, j]), indcond

class _ConstantPredictor(BaseEstimator):
    """ Predicts same labels as trained """ 
    def fit(self, X, y):
        self.y_ = y
        return self

    def predict(self, X):
        check_is_fitted(self, 'y_')
        return np.repeat(self.y_, X.shape[0])

    def decision_function(self, X):
        check_is_fitted(self, 'y_')
        return np.repeat(self.y_, X.shape[0])

    def predict_proba(self, X):
        check_is_fitted(self, 'y_')
        return np.repeat([np.hstack([1 - self.y_, self.y_])],
                         X.shape[0], axis=0)

class DistOneVsRestClassifier(OneVsRestClassifier):
    """
    Same as sklearn `OneVsRestClassifier` but with distributed
    training using spark. Additionally implements flexible
    ``predict_proba`` method with custom `norm` input
    designating the normalization method used after individual
    predictions are made.

    Args:
        estimator (sklearn estimator): An estimator object implementing fit and one of
            decision_function or predict_proba.
        sc (sparkContext): Spark context for spark broadcasting and rdd operations.
        norm (string): default None, Normalization method for predict_proba.
        partitions (int or 'auto'): default 'auto'
            Number of partitions to use for parallelization of parameter
            search space. Integer values or None will be used directly for `numSlices`,
            while 'auto' will set `numSlices` to the number required fits.
        max_negatives (int or float): default None
            Maximum number of negative records allowed for each binary
            estimator. Use int for hard maximum, or float for percentage
            of total negatives.
        random_state (int): default None
            Random state for limiting negatives (if max_negatives is not None).
        method (str): 'ratio' or 'multiplier'
            Method used to calculate true maximum number of negatives.
        n_splits (int): default 1
            Dials the number of splits for broadcasting
            X during fitting. Use values higher than 1 for large X.
        mlb_override (bool): pass over mlb step; this assumes
            that input `y` to `fit` is already in sparse (one-hot-encoded)
            format
        verbose (bool): print status messages
        **kwargs: Keyword arguments to be passed to `OneVsRestClassifier`.
    """
    def __init__(self, estimator, sc=None, norm=None, partitions='auto',
            max_negatives=None, random_state=None, method="ratio",
            n_splits=1, mlb_override=False, verbose=False, **kwargs):
        OneVsRestClassifier.__init__(
            self, estimator, **kwargs)
        self.norm = norm
        self.sc = sc
        self.partitions = partitions
        self.max_negatives = max_negatives
        self.random_state = random_state
        self.method = method
        self.n_splits = n_splits
        self.mlb_override = mlb_override
        self.verbose = verbose

    def fit(self, X, y, **fit_params):
        """
        Fit underlying estimators. Parallelize fit operation using spark.

        Args:
            X (array-like, shape = [n_samples, n_features]): input data
            y (array-like, shape = [n_samples, ], [n_samples, n_classes]): multi-class targets
            **fit_params (dict of string -> object): parameters passed 
                to the ``fit`` method of the estimator
        """
        _check_estimator(self, verbose=self.verbose)

        if (not self.mlb_override and not hasattr(y[0], '__array__') 
                and isinstance(y[0], Sequence)
                and not isinstance(y[0], str)):
            self.mlb = MultiLabelBinarizer()
            y = self.mlb.fit_transform(y)

        if isinstance(X, pd.DataFrame):
            X.index = list(range(len(X)))

        self.label_binarizer_ = LabelBinarizer(sparse_output=True)
        self.label_binarizer_.fit(y)
        self.classes_ = self.label_binarizer_.classes_
        self._fit(X, y, **fit_params)
        del self.sc
        if hasattr(self.estimator, "sc"):
            del self.estimator.sc
        return self

    def _fit(self, X, y, **fit_params):
        Y = self.label_binarizer_.transform(y)
        Y = Y.tocsc()
        max_negatives = self.max_negatives
        random_state = self.random_state
        n_splits = self.n_splits
        method = self.method
        estimator = _clone(self.estimator)
        if self.sc is None:
            models_ = Parallel(n_jobs=self.n_jobs)(
                delayed(_fit_binary)(
                    estimator, X, x[1], fit_params, 
                    classes=["not %s" % x[0], x[0]],
                    max_negatives=max_negatives, 
                    random_state=random_state, method=method)
                for x in list(zip(self.classes_, list(col.toarray().ravel() for col in Y.T))))
        else:
            X = _split_X(X, n_splits, self.sc)
            partitions = _parse_partitions(self.partitions, len(self.classes_))
            estimator = self.sc.broadcast(self.estimator)
            columns = self.sc.parallelize(
                list(zip(self.classes_, list(col.toarray().ravel() for col in Y.T))),
                numSlices=partitions)
            models_ = columns.map(lambda x: _fit_binary(
                estimator, X, x[1], fit_params, classes=["not %s" % x[0], x[0]],
                max_negatives=max_negatives, random_state=random_state, method=method)).collect()
        estimators_ = [x[0] for x in models_]
        classes_ = [x[1] for x in models_]
        self.estimators_ = list([estimators_[classes_.index(x)] for x in self.classes_])
        return self

    def predict_proba(self, X):
        """
        Probability estimates.
        The returned estimates for all classes are ordered by label of classes.

        Args:
            X (array-like, shape = [n_samples, n_features]): input data

        Returns:
            T (array-like, shape = [n_samples, n_classes]): returns the probability 
                of the sample for each class in the model, where classes are 
                ordered as they are in self.classes_
        """
        probs = []
        for index in range(len(self.estimators_)):
            probs.append(self.estimators_[index].predict_proba(X)[:,1])
        out = np.array([
            [probs[y][index] for y in range(len(self.estimators_))]
            for index in range(len(probs[0]))])
        if self.norm:
            return normalize(out, norm=self.norm)
        else:
            return out

class DistOneVsOneClassifier(OneVsOneClassifier):
    """
    Same as sklearn `OneVsOneClassifier` but with distributed
    training using spark.

    Args:
        estimator (sklearn estimator): An estimator object implementing fit and one of
            decision_function or predict_proba.
        sc (sparkContext): Spark context for spark broadcasting and rdd operations.
        partitions (int or 'auto'): default 'auto'
            Number of partitions to use for parallelization of parameter
            search space. Integer values or None will be used directly for `numSlices`,
            while 'auto' will set `numSlices` to the number required fits.
        verbose (bool): print status messages
        **kwargs: Keyword arguments to be passed to `OneVsOneClassifier`.
    """
    def __init__(self, estimator, sc=None, partitions='auto', verbose=False, **kwargs):
        OneVsOneClassifier.__init__(
            self, estimator, **kwargs)
        self.sc = sc
        self.partitions = partitions
        self.verbose = verbose

    def fit(self, X, y, **fit_params):
        """
        Fit underlying estimators. Parallelize fit operation using spark.

        Args:
            X (array-like, shape = [n_samples, n_features]): intput
                data
            y (array-like, shape = [n_samples]): Multi-class targets.
            **fit_params (dict of string -> object): parameters passed 
                to the ``fit`` method of the estimator
        """
        _check_estimator(self, verbose=self.verbose)

        X, y = check_X_y(X, y, accept_sparse=['csr', 'csc'])
        check_classification_targets(y)

        self.classes_ = np.unique(y)
        if len(self.classes_) == 1:
            raise ValueError("OneVsOneClassifier can not be fit when only one"
                             " class is present.")
        n_classes = self.classes_.shape[0]
        class_list = []
        count = -1
        for i in range(n_classes):
            for j in range(i + 1, n_classes):
                count += 1
                class_list.append([count, self.classes_[i], self.classes_[j]])
        estimator = _clone(self.estimator)

        if self.sc is None:
            estimators_indices = list(zip(*(Parallel(n_jobs=self.n_jobs)(
                delayed(_fit_ovo_binary)
                (self.estimator, X, y, self.classes_[i], self.classes_[j], fit_params)
                for i in range(n_classes) for j in range(i + 1, n_classes)))))
            self.estimators_ = [x[0] for x in estimators_indices[0]]
            self.pairwise_indices_ = (
                estimators_indices[1] if self._pairwise else None)
        else:
            estimator = self.sc.broadcast(self.estimator)
            partitions = _parse_partitions(self.partitions, len(class_list))
            estimators_indices = self.sc.parallelize(
                    class_list,
                    numSlices=partitions).groupBy(lambda x: x[0]).map(
                    lambda x: [x[0], _fit_ovo_binary(estimator, X, y, list(x[1])[0][1], list(x[1])[0][2], fit_params)]).collect()

            estimators_indices = [estimators_indices[i][1] for i in np.argsort([x[0] for x in estimators_indices])]
            self.estimators_ = [x[0][0] for x in estimators_indices]
            try:
                self.pairwise_indices_ = (
                    [x[1] for x in estimators_indices] if self._pairwise else None)
            except AttributeError:
                self.pairwise_indices_ = None
        del self.sc
        if hasattr(self.estimator, "sc"):
            del self.estimator.sc
        return self



"""
Test defaults imports
"""

try:
    from skdist.distribute import _defaults
    from skdist.distribute._defaults import *
    _import_error = None
except Exception as e:
    _import_error = e

def test_import_defaults():
    assert _import_error == None

def test_tokenizer():
    assert tokenizer(5) == 5

def test_dict():
    func = _defaults.dict_encoder("a")
    assert isinstance(func, list)

def test_numeric():
    func = _defaults.numeric_encoder("a")
    assert isinstance(func, list)

def test_onehot():
    func = _defaults.onehot_encoder("a")
    assert isinstance(func, list)

def test_multihot():
    func = _defaults.multihot_encoder("a")
    assert isinstance(func, list)

def test_small():
    for key, val in _defaults._default_encoders["small"].items():
        func = val("c")
    assert isinstance(func, list)

def test_medium():
    for key, val in _defaults._default_encoders["medium"].items():
        func = val("c")
    assert isinstance(func, list)

def test_large():
    for key, val in _defaults._default_encoders["large"].items():
        func = val("c")
    assert isinstance(func, list)
     



"""
Feature selection models
"""

import numpy as np

from sklearn.base import (
    BaseEstimator, ClassifierMixin, 
    is_classifier
    )
from sklearn.utils.metaestimators import if_delegate_has_method
from sklearn.utils.validation import check_is_fitted 
from sklearn.utils import check_X_y, safe_sqr
from sklearn.model_selection import check_cv
from sklearn.metrics.scorer import check_scoring
from joblib import Parallel, delayed
from itertools import product
from scipy.sparse import issparse

from .base import _parse_partitions, _clone
from .utils import _safe_split

__all__ = ["DistFeatureEliminator"]

def _drop_col(X, index):
    """ Drop index columns from numpy array or sparse matrix """
    if issparse(X):
        cols = np.arange(X.shape[1])
        cols_to_keep = np.where(np.logical_not(np.in1d(cols, index)))[0]
        return X[:, cols_to_keep]
    else:
        return np.delete(X, index, axis=1)

def _fit_and_score_one(index, estimator, X, y, scorer, train, test, verbose, fit_params):
    """ Fit and score an estimator with one feature left out """
    estimator_ = _clone(estimator)
    X_train, y_train = _safe_split(estimator, _drop_col(X, index), y, train)
    X_test, y_test = _safe_split(estimator, _drop_col(X, index), y, test, train)
    estimator_.fit(X_train, y_train, **fit_params)
    return scorer(estimator_, X_test, y_test)

def _divide_chunks(l, n): 
    """ Divide iterable into n-sized chunks """
    for i in range(0, len(l), n):  
        yield l[i:i + n] 

class DistFeatureEliminator(BaseEstimator, ClassifierMixin):
    """
    Similar to ``sklearn.feature_selection.RFECV`` but with distributed
    feature set removal and cross validation. Fits the base estimator on 
    the input training data once for each left out feature sets according 
    to the `step` and the scoring of the features according to an initial 
    estimator fit with all features. Uses cross validation to score each 
    feature set. Results in the fitted base estimator with the best feature set.

    Roughly approximates `RFECV` particularly as `step` approaches 1 and 
    `min_features_to_select` approaches 0. The value-add is in the 
    embarrassingly parallel nature of the `DistFeatureEliminator` algorithm.
    
    Args:
        estimator (estimator object):
            This is assumed to implement the scikit-learn estimator interface.
            Either estimator needs to provide a ``score`` function,
            or ``scoring`` must be passed.
        sc (sparkContext): Spark context for spark broadcasting and rdd operations.
        partitions (int or 'auto'): default 'auto'
            Number of partitions to use for parallelization of parameter
            search space. Integer values or None will be used directly for `numSlices`,
            while 'auto' will set `numSlices` to the number required fits.
        min_features_to_select (int): Minimum number of features to keep after elimination.
        step (int or float): If greater than or equal to 1, then `step` 
            corresponds to the (integer) number of features to remove at each iteration.
            If within (0.0, 1.0), then `step` corresponds to the percentage
            (rounded down) of features to remove at each iteration.
        cv (int or cv object): Determines the cross-validation splitting strategy.
        scoring (str or callable): Scoring function or callable to evaluate 
            predictions on the training set for feature elimination.
        verbose (int or bool): Controls the verbosity: the higher, the more messages.
        n_jobs (int): Number of jobs to run in parallel if using joblib.
        pre_dispatch (bool): Controls number of jobs dispatched if using joblib.
        mask (bool): Whether or not to apply feature selection mask to input data
            before making predictions.
    """
    def __init__(self, 
                 estimator, 
                 sc=None,
                 partitions='auto',
                 min_features_to_select=None, 
                 step=1,
                 cv=5, 
                 scoring=None, 
                 verbose=False, 
                 n_jobs=None, 
                 pre_dispatch=None, 
                 mask=True):
        self.estimator = estimator
        self.sc = sc
        self.partitions = partitions
        self.min_features_to_select = min_features_to_select
        self.step = step
        self.cv = cv
        self.scoring = scoring
        self.verbose = verbose
        self.n_jobs = n_jobs
        self.pre_dispatch = pre_dispatch
        self.mask = mask
        
    def fit(self, X, y=None, groups=None, **fit_params):
        """ 
        Apply feature elimination routine, ultimately fitting 
        estimator on the best feature set.

        Args:
            X (array-like, shape = [n_samples, n_features]): input data
            y (array-like, shape = [n_samples, ], [n_samples, n_classes]): targets
            groups (array-like): group labels for the samples used while 
                splitting the dataset into train/test set
            **fit_params (dict of string -> object): parameters passed
                to the `fit` method of the estimator
        """
        X, y = check_X_y(X, y, "csr", ensure_min_features=2)
        cv = check_cv(self.cv, y, is_classifier(self.estimator))
        scorer = check_scoring(self.estimator, scoring=self.scoring)

        n_features = X.shape[1]
        if self.min_features_to_select is None:
            min_features_to_select = n_features // 2
        else:
            min_features_to_select = self.min_features_to_select

        if 0.0 < self.step < 1.0:
            step = int(max(1, self.step * n_features))
        else:
            step = int(self.step)
        if step <= 0:
            raise ValueError("Step must be >0")

        initial_estimator = _clone(self.estimator)
        initial_estimator.fit(X, y, **fit_params)
        if hasattr(initial_estimator, 'coef_'):
            coefs = initial_estimator.coef_
        else:
            coefs = getattr(initial_estimator, 'feature_importances_', None)
        if coefs is None:
            raise RuntimeError('The classifier does not expose '
                                '"coef_" or "feature_importances_" '
                                'attributes')
        if coefs.ndim > 1:
            ranks = np.argsort(safe_sqr(coefs).sum(axis=0))
        else:
            ranks = np.argsort(safe_sqr(coefs))
        ranks = np.ravel(ranks)[:(n_features - min_features_to_select)]

        this_step = 0
        features_to_remove = [np.array([])]
        while this_step < (n_features - min_features_to_select):
            this_step += step
            features_to_remove.append(ranks[:this_step])

        cv_splits_ = list(cv.split(X,y,groups))
        fit_sets = list(product(features_to_remove, cv_splits_))
        base_estimator = _clone(self.estimator)
        if not self.sc:
            parallel = Parallel(
                n_jobs=self.n_jobs, verbose=self.verbose, 
                pre_dispatch=self.pre_dispatch
                )
            scores = parallel(
                delayed(_fit_and_score_one)(
                    index, _clone(base_estimator), X, y, 
                    scorer, train, test, self.verbose,
                    fit_params)
                for index, (train, test) in fit_sets)
            score_sets = _divide_chunks(list(scores), len(cv_splits_))
        else:
            base_estimator_ = self.sc.broadcast(base_estimator)
            partitions = _parse_partitions(self.partitions, len(fit_sets))
            verbose = self.verbose
            scores = (
                self.sc.parallelize(fit_sets, numSlices=partitions)
                .map(lambda x: [x[0], _fit_and_score_one(
                    x[0], _clone(base_estimator), X, y, scorer, 
                    x[1][0], x[1][1], verbose, fit_params)]).collect()
                )
            score_sets = []
            for feat_set in features_to_remove:
                this_set = []
                for row in scores:
                    if (feat_set.shape == row[0].shape) and np.allclose(feat_set, row[0]):
                        this_set.append(row[1])
                score_sets.append(this_set)
            
        self.scores_ = []
        for score_set in score_sets:
            this_score = np.mean(score_set)
            self.scores_.append(this_score)
            
        best_set_ = np.argmax(self.scores_)
        self.best_score_ = self.scores_[best_set_]
        self.best_features_ = np.delete(range(n_features), features_to_remove[best_set_])
        self.best_estimator_ = _clone(self.estimator)
        self.best_estimator_.fit(X[:, self.best_features_], y, **fit_params)
        self.n_features_ = len(self.best_features_)
        
        del self.sc
        return self   
    
    def _apply_mask(self, X):
        self._check_is_fitted('best_features_')
        if self.mask:
            return X[:, self.best_features_]
        else:
            return X
    
    def _check_is_fitted(self, method_name):
        check_is_fitted(self, method_name)
    
    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def predict(self, X):
        self._check_is_fitted('predict')
        return self.best_estimator_.predict(self._apply_mask(X))

    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def predict_proba(self, X):
        self._check_is_fitted('predict_proba')
        return self.best_estimator_.predict_proba(self._apply_mask(X))

    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def predict_log_proba(self, X):
        self._check_is_fitted('predict_log_proba')
        return self.best_estimator_.predict_log_proba(self._apply_mask(X))

    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def decision_function(self, X):
        self._check_is_fitted('decision_function')
        return self.best_estimator_.decision_function(self._apply_mask(X))

    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def transform(self, X):
        self._check_is_fitted('transform')
        return self.best_estimator_.transform(self._apply_mask(X))
    
    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def score(self, X, y):
        self._check_is_fitted('score')
        return self.best_estimator_.score(self._apply_mask(X), y)
        
    @property
    def classes_(self):
        self._check_is_fitted("classes_")
        return self.best_estimator_.classes_




import pandas as pd

import pyspark
from pyspark.sql import SparkSession, functions as F

from sklearn.datasets import (
    load_breast_cancer,
    load_digits
)
from sklearn.model_selection import train_test_split
from sklearn.linear_model import LogisticRegression

from skdist.distribute.multiclass import DistOneVsRestClassifier
from skdist.distribute.search import DistGridSearchCV
from skdist.distribute.ensemble import DistRandomForestClassifier
from skdist.distribute.predict import get_prediction_udf

def test_spark_session_dataframe(spark_session):
    test_df = spark_session.createDataFrame([[1, 3], [2, 4]], "a: int, b: int")

    assert type(test_df) == pyspark.sql.dataframe.DataFrame
    assert test_df.count() == 2

def test_spark_session_sql(spark_session):
    test_df = spark_session.createDataFrame([[1, 3], [2, 4]], "a: int, b: int")
    test_df.createOrReplaceTempView('test')

    test_filtered_df = spark_session.sql('SELECT a, b from test where a > 1')
    assert test_filtered_df.count() == 1

def test_ensemble(spark_session):
    sc = spark_session.sparkContext

    test_size = 0.2
    max_depth = None
    n_estimators = 100

    # load sample data (binary target)
    data = load_breast_cancer()
    X = data["data"]
    y = data["target"]
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=test_size, random_state=10
    )

    ### distributed random forest
    model = DistRandomForestClassifier(
        n_estimators=n_estimators,
        max_depth=max_depth, sc=sc,
    )
    # distributed fitting with spark
    model.fit(X_train, y_train)
    # predictions on the driver
    preds = model.predict(X_test)

    assert preds.shape == y_test.shape

def test_search(spark_session):
    sc = spark_session.sparkContext

    # sklearn variables
    Cs = [0.001, 0.01, 0.1, 1.0, 10.0, 100.0]
    cv = 5
    test_size = 0.2
    scoring = "roc_auc"
    solver = "liblinear"

    # load sample data (binary target)
    data = load_breast_cancer()
    X = data["data"]
    y = data["target"]
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=test_size, random_state=10
    )

    ### distributed grid search
    model = DistGridSearchCV(
        LogisticRegression(solver=solver),
        dict(C=Cs), sc, cv=cv, scoring=scoring
    )
    # distributed fitting with spark
    model.fit(X_train, y_train)
    # predictions on the driver
    preds = model.predict(X_test)

    assert preds.shape == y_test.shape

def test_multiclass(spark_session):
    sc = spark_session.sparkContext

    # variables
    solver = "liblinear"
    test_size = 0.2

    # load sample data (binary target)
    data = load_digits()
    X = data["data"]
    y = data["target"]
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=test_size, random_state=10
    )

    ### distributed one vs rest
    model = DistOneVsRestClassifier(LogisticRegression(solver=solver), sc)
    # distributed fitting with spark
    model.fit(X_train, y_train)
    # predictions on the driver
    preds = model.predict(X_test)

    assert preds.shape == y_test.shape

def test_predict(spark_session):
    sc = spark_session.sparkContext

    # simple 2-D numpy features
    data = load_digits()
    X = data["data"]
    y = data["target"]
    model = LogisticRegression(
        solver="liblinear",
        multi_class="auto"
    )
    model.fit(X, y)

    # get UDFs with default 'numpy' feature types
    predict = get_prediction_udf(model, method="predict")
    predict_proba = get_prediction_udf(model, method="predict_proba")

    # create PySpark DataFrame from features
    pdf = pd.DataFrame(X)
    sdf = spark_session.createDataFrame(pdf)
    cols = [F.col(str(c)) for c in sdf.columns]

    # apply predict UDFs and select prediction output
    prediction_df = (
        sdf
            .withColumn("scores", predict_proba(*cols))
            .withColumn("preds", predict(*cols))
            .select("preds", "scores")
    )
    assert prediction_df.count() == X.shape[0]


"""
Distributed ensemble meta-estimators
"""

import numpy as np
import pandas as pd

from joblib import Parallel, delayed
from scipy.sparse import hstack, issparse
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.ensemble.forest import (
    BaseForest, ExtraTreesClassifier,
    ExtraTreesRegressor, ForestClassifier,
    ForestRegressor, RandomForestClassifier,
    RandomForestRegressor
    )
from sklearn.tree import ExtraTreeRegressor
from sklearn.preprocessing import OneHotEncoder
from sklearn.model_selection import KFold
from sklearn.tree.tree import DOUBLE, DTYPE
from sklearn.utils import check_array, check_random_state
from sklearn.utils.validation import check_is_fitted

from .validation import _check_estimator
from .base import (
    _clone, _get_value, _parse_partitions
    )

__all__ = [
    "DistRandomForestClassifier",
    "DistRandomForestRegressor",
    "DistExtraTreesClassifier",
    "DistExtraTreesRegressor",
    "DistRandomTreesEmbedding"
]

MAX_RAND_SEED = np.iinfo(np.int32).max

def _set_random_states(estimator, random_state=None):
    """ Sets fixed random_state parameters for an estimator """
    to_set = {}
    for key in sorted(estimator.get_params(deep=True)):
        if key == 'random_state' or key.endswith('__random_state'):
            to_set[key] = random_state
    if to_set:
        estimator.set_params(**to_set)

def _generate_sample_indices(random_state, n_samples):
    """ Private function used to _parallel_build_trees function """
    random_instance = check_random_state(random_state)
    sample_indices = random_instance.randint(0, n_samples, n_samples)
    return sample_indices

def _make_estimator(base_estimator, estimator_params, params={}, random_state=None):
    """ Make and configure a copy of base_estimator """
    estimator = _clone(base_estimator)
    estimator.set_params(**dict((p, params[p])
                                for p in estimator_params))

    if random_state is not None:
        _set_random_states(estimator, random_state)
    return estimator

def _build_trees(base_estimator, estimator_params, params, X, y, sample_weight,
                          tree_state, n_trees, verbose=0, class_weight=None,
                          bootstrap=False):
    """ Fit a single tree in parallel """
    tree = _make_estimator(
        _get_value(base_estimator), estimator_params,
        params=params, random_state=tree_state
        )
    if bootstrap:
        n_samples = X.shape[0]
        if sample_weight is None:
            curr_sample_weight = np.ones((n_samples,), dtype=np.float64)
        else:
            curr_sample_weight = sample_weight.copy()

        indices = _generate_sample_indices(tree_state, n_samples)
        sample_counts = np.bincount(indices, minlength=n_samples)
        curr_sample_weight *= sample_counts

        if class_weight == 'subsample':
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', DeprecationWarning)
                curr_sample_weight *= compute_sample_weight('auto', y, indices)
        elif class_weight == 'balanced_subsample':
            curr_sample_weight *= compute_sample_weight('balanced', y, indices)

        tree.fit(X, y, sample_weight=curr_sample_weight, check_input=False)
    else:
        tree.fit(X, y, sample_weight=sample_weight, check_input=False)
    return tree

def get_single_oof(clf, X, y, train_index, test_index):
    """
    Fit on the data specified by the train_index and predict_proba on the
    test index
    """
    if isinstance(X, (pd.DataFrame, pd.Series)):
        X_train = X.iloc[train_index]
        X_test = X.iloc[test_index]
    else:
        X_train = X[train_index]
        X_test = X[test_index]

    y_train = y[train_index]

    clf.fit(X_train, y_train)
    return test_index, clf.predict_proba(X_test)

def get_oof(clf, X, y, n_splits=5):
    """ Fit the classifier to the data and make predictions out of fold """
    kfold = KFold(n_splits=n_splits)

    oof_train = np.zeros((y.shape[0], len(np.unique(y))))

    for train_index, test_index in kfold.split(X):
        if isinstance(X, (pd.DataFrame, pd.Series)):
            X_train = X.iloc[train_index]
            X_test = X.iloc[test_index]
        else:
            X_train = X[train_index]
            X_test = X[test_index]

        y_train = y[train_index]

        clf.fit(X_train, y_train)

        oof_train[test_index] = clf.predict_proba(X_test)

    clf.fit(X, y)
    return clf, oof_train

class DistBaseForest(BaseForest):
    """
    Same as sklearn `BaseForest` but with distributed
    training using spark.

    Args:
        base_estimator (sklearn estimator): An estimator object implementing fit and one of
            decision_function or predict_proba.
        sc (sparkContext): Spark context for spark broadcasting and rdd operations.
        partitions (int or 'auto'): default 'auto'
            Number of partitions to use for parallelization of parameter
            search space. Integer values or None will be used directly for `numSlices`,
            while 'auto' will set `numSlices` to the number required fits.
        **kwargs: Keyword arguments to be passed to `BaseForest`.
    """
    def __init__(self, base_estimator, sc=None, partitions='auto', **kwargs):
        BaseForest.__init__(
            self, base_estimator, **kwargs)
        self.sc = sc
        self.partitions = partitions
        for key, val in kwargs.items():
            setattr(self, key, val)

    def fit(self, X, y, sample_weight=None):
        """
        Build a forest of trees from the training set (X, y). Parallelize fit
        operation using spark.

        Args:
            X (array-like or sparse matrix of shape = [n_samples, n_features])
                The training input samples. Internally, its dtype will be converted to
                ``dtype=np.float32``. If a sparse matrix is provided, it will be
                converted into a sparse ``csc_matrix``.
            y (array-like, shape = [n_samples] or [n_samples, n_outputs])
                The target values (class labels in classification, real numbers in
                regression).
            sample_weight (array-like, shape = [n_samples] or None)
                Sample weights. If None, then samples are equally weighted. Splits
                that would create child nodes with net zero or negative weight are
                ignored while searching for a split in each node. In the case of
                classification, splits are also ignored if they would result in any
                single class carrying a negative weight in either child node.
        """
        _check_estimator(self, verbose=self.verbose)        

        # Validate or convert input data
        X = check_array(X, accept_sparse="csc", dtype=DTYPE)
        y = check_array(y, accept_sparse='csc', ensure_2d=False, dtype=None)
        if sample_weight is not None:
            sample_weight = check_array(sample_weight, ensure_2d=False)
        if issparse(X):
            # Pre-sort indices to avoid that each individual tree of the
            # ensemble sorts the indices.
            X.sort_indices()

        # Remap output
        n_samples, self.n_features_ = X.shape

        y = np.atleast_1d(y)
        if y.ndim == 2 and y.shape[1] == 1:
            warn("A column-vector y was passed when a 1d array was"
                 " expected. Please change the shape of y to "
                 "(n_samples,), for example using ravel().",
                 DataConversionWarning, stacklevel=2)

        if y.ndim == 1:
            # reshape is necessary to preserve the data contiguity against vs
            # [:, np.newaxis] that does not.
            y = np.reshape(y, (-1, 1))

        self.n_outputs_ = y.shape[1]

        y, expanded_class_weight = self._validate_y_class_weight(y)

        if getattr(y, "dtype", None) != DOUBLE or not y.flags.contiguous:
            y = np.ascontiguousarray(y, dtype=DOUBLE)

        if expanded_class_weight is not None:
            if sample_weight is not None:
                sample_weight = sample_weight * expanded_class_weight
            else:
                sample_weight = expanded_class_weight

        # Check parameters
        self._validate_estimator()

        if not self.bootstrap and self.oob_score:
            raise ValueError("Out of bag estimation only available"
                             " if bootstrap=True")

        random_state = check_random_state(self.random_state)

        if not self.warm_start or not hasattr(self, "estimators_"):
            # Free allocated memory, if any
            self.estimators_ = []

        n_more_estimators = self.n_estimators - len(self.estimators_)

        if n_more_estimators < 0:
            raise ValueError('n_estimators=%d must be larger or equal to '
                             'len(estimators_)=%d when warm_start==True'
                             % (self.n_estimators, len(self.estimators_)))

        elif n_more_estimators == 0:
            warn("Warm-start fitting without increasing n_estimators does not "
                 "fit new trees.")
        else:
            if self.warm_start and len(self.estimators_) > 0:
                # We draw from the random state to get the random state we
                # would have got if we hadn't used a warm_start.
                random_state.randint(MAX_RAND_SEED, size=len(self.estimators_))

            estimator_params = self.estimator_params
            verbose = self.verbose
            class_weight = self.class_weight
            bootstrap = self.bootstrap
            states = list(random_state.randint(MAX_RAND_SEED, size=n_more_estimators))
            params = {}
            for p in self.estimator_params:
                params[p] = getattr(self, p)

            if self.sc is None:
                base_estimator = self.base_estimator_
                trees = Parallel(n_jobs=self.n_jobs)(
                    delayed(_build_trees)(
                        base_estimator, estimator_params, 
                        params, X, y, sample_weight,
                        state, n_more_estimators, verbose=verbose,
                        class_weight=class_weight, 
                        bootstrap=bootstrap)
                    for state in states)
            else:
                base_estimator = self.sc.broadcast(self.base_estimator_)
                partitions = _parse_partitions(self.partitions, n_more_estimators)
                trees = self.sc.parallelize(
                    states,
                    numSlices=partitions).map(
                    lambda x: _build_trees(
                        base_estimator, estimator_params, params, X, y, sample_weight,
                        x, n_more_estimators, verbose=verbose,
                        class_weight=class_weight, bootstrap=bootstrap)).collect()

            # Collect newly grown trees
            self.estimators_.extend(trees)

        if self.oob_score:
            self._set_oob_score(X, y)

        # Decapsulate classes_ attributes
        if hasattr(self, "classes_") and self.n_outputs_ == 1:
            self.n_classes_ = self.n_classes_[0]
            self.classes_ = self.classes_[0]

        del self.sc
        return self

    def _set_oob_score(self, X, y):
        """ Calculate out of bag predictions and score """
        pass

class DistForestClassifier(DistBaseForest, ForestClassifier):
    """
    Same as sklearn `ForestClassifier` but with distributed
    training using spark.

    Args:
        base_estimator (sklearn estimator): An estimator object implementing fit and one of
            decision_function or predict_proba.
        sc (sparkContext): Spark context for spark broadcasting and rdd operations.
        partitions (int or 'auto'): default 'auto'
            Number of partitions to use for parallelization of parameter
            search space. Integer values or None will be used directly for `numSlices`,
            while 'auto' will set `numSlices` to the number required fits.
        **kwargs: Keyword arguments to be passed to `ForestClassifier`.
    """
    def __init__(self, base_estimator, sc=None, partitions='auto', **kwargs):
        ForestClassifier.__init__(
            self, base_estimator, **kwargs)
        self.sc = sc
        self.partitions = partitions

class DistRandomForestClassifier(DistForestClassifier, RandomForestClassifier):
    """
    Same as sklearn `RandomForestClassifier` but with distributed
    training using spark.

    Args:
        sc (sparkContext): Spark context for spark broadcasting and rdd operations.
        partitions (int or 'auto'): default 'auto'
            Number of partitions to use for parallelization of parameter
            search space. Integer values or None will be used directly for `numSlices`,
            while 'auto' will set `numSlices` to the number required fits.
    """
    def __init__(self, 
                 sc=None, 
                 partitions='auto', 
                 n_estimators=100,
                 criterion="gini",
                 max_depth=None,
                 min_samples_split=2,
                 min_samples_leaf=1,
                 min_weight_fraction_leaf=0.,
                 max_features="auto",
                 max_leaf_nodes=None,
                 min_impurity_decrease=0.,
                 min_impurity_split=None,
                 bootstrap=True,
                 oob_score=False,
                 n_jobs=None,
                 random_state=None,
                 verbose=0,
                 warm_start=False,
                 class_weight=None):
        RandomForestClassifier.__init__(
            self, 
            n_estimators=n_estimators,
            criterion=criterion,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_features=max_features,
            max_leaf_nodes=max_leaf_nodes,
            min_impurity_decrease=min_impurity_decrease,
            min_impurity_split=min_impurity_split,
            bootstrap=bootstrap,
            oob_score=oob_score,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            warm_start=warm_start,
            class_weight=class_weight
            )
        self.sc = sc
        self.partitions = partitions

class DistExtraTreesClassifier(DistForestClassifier, ExtraTreesClassifier):
    """
    Same as sklearn `ExtraTreesClassifier` but with distributed
    training using spark.

    Args:
        sc (sparkContext): Spark context for spark broadcasting and rdd operations.
        partitions (int or 'auto'): default 'auto'
            Number of partitions to use for parallelization of parameter
            search space. Integer values or None will be used directly for `numSlices`,
            while 'auto' will set `numSlices` to the number required fits.
    """
    def __init__(self, 
                 sc=None, 
                 partitions='auto', 
                 n_estimators=100,
                 criterion="gini",
                 max_depth=None,
                 min_samples_split=2,
                 min_samples_leaf=1,
                 min_weight_fraction_leaf=0.,
                 max_features="auto",
                 max_leaf_nodes=None,
                 min_impurity_decrease=0.,
                 min_impurity_split=None,
                 bootstrap=False,
                 oob_score=False,
                 n_jobs=None,
                 random_state=None,
                 verbose=0,
                 warm_start=False,
                 class_weight=None):
        ExtraTreesClassifier.__init__(
            self,
            n_estimators=n_estimators,
            criterion=criterion,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_features=max_features,
            max_leaf_nodes=max_leaf_nodes,
            min_impurity_decrease=min_impurity_decrease,
            min_impurity_split=min_impurity_split,
            bootstrap=bootstrap,
            oob_score=oob_score,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            warm_start=warm_start,
            class_weight=class_weight
            )
        self.sc = sc
        self.partitions = partitions

class DistForestRegressor(DistBaseForest, ForestRegressor):
    """
    Same as sklearn `ForestRegressor` but with distributed
    training using spark.

    Args:
        base_estimator (sklearn estimator): An estimator object implementing fit and one of
            decision_function or predict_proba.
        sc (sparkContext): Spark context for spark broadcasting and rdd operations.
        partitions (int or 'auto'): default 'auto'
            Number of partitions to use for parallelization of parameter
            search space. Integer values or None will be used directly for `numSlices`,
            while 'auto' will set `numSlices` to the number required fits.
        **kwargs: Keyword arguments to be passed to `ForestRegressor`.
    """
    def __init__(self, base_estimator, sc=None, partitions='auto', **kwargs):
        ForestRegressor.__init__(
            self, base_estimator, **kwargs)
        self.sc = sc
        self.partitions = partitions

class DistRandomForestRegressor(DistForestRegressor, RandomForestRegressor):
    """
    Same as sklearn `RandomForestRegressor` but with distributed
    training using spark.

    Args:
        sc (sparkContext): Spark context for spark broadcasting and rdd operations.
        partitions (int or 'auto'): default 'auto'
            Number of partitions to use for parallelization of parameter
            search space. Integer values or None will be used directly for `numSlices`,
            while 'auto' will set `numSlices` to the number required fits.
    """
    def __init__(self,
                 sc=None,
                 partitions='auto',
                 n_estimators=100,
                 criterion="mse",
                 max_depth=None,
                 min_samples_split=2,
                 min_samples_leaf=1,
                 min_weight_fraction_leaf=0.,
                 max_features="auto",
                 max_leaf_nodes=None,
                 min_impurity_decrease=0.,
                 min_impurity_split=None,
                 bootstrap=True,
                 oob_score=False,
                 n_jobs=None,
                 random_state=None,
                 verbose=0,
                 warm_start=False):
        RandomForestRegressor.__init__(
            self,
            n_estimators=n_estimators,
            criterion=criterion,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_features=max_features,
            max_leaf_nodes=max_leaf_nodes,
            min_impurity_decrease=min_impurity_decrease,
            min_impurity_split=min_impurity_split,
            bootstrap=bootstrap,
            oob_score=oob_score,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            warm_start=warm_start
            )
        self.sc = sc
        self.partitions = partitions

class DistExtraTreesRegressor(DistForestRegressor, ExtraTreesRegressor):
    """
    Same as sklearn `ExtraTreesRegressor` but with distributed
    training using spark.

    Args:
        sc (sparkContext): Spark context for spark broadcasting and rdd operations.
        partitions (int or 'auto'): default 'auto'
            Number of partitions to use for parallelization of parameter
            search space. Integer values or None will be used directly for `numSlices`,
            while 'auto' will set `numSlices` to the number required fits.
    """
    def __init__(self,
                 sc=None,
                 partitions='auto',
                 n_estimators=100,
                 criterion="mse",
                 max_depth=None,
                 min_samples_split=2,
                 min_samples_leaf=1,
                 min_weight_fraction_leaf=0.,
                 max_features="auto",
                 max_leaf_nodes=None,
                 min_impurity_decrease=0.,
                 min_impurity_split=None,
                 bootstrap=False,
                 oob_score=False,
                 n_jobs=None,
                 random_state=None,
                 verbose=0,
                 warm_start=False):
        ExtraTreesRegressor.__init__(
            self,
            n_estimators=n_estimators,
            criterion=criterion,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_features=max_features,
            max_leaf_nodes=max_leaf_nodes,
            min_impurity_decrease=min_impurity_decrease,
            min_impurity_split=min_impurity_split,
            bootstrap=bootstrap,
            oob_score=oob_score,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            warm_start=warm_start
            )
        self.sc = sc
        self.partitions = partitions

class DistRandomTreesEmbedding(DistBaseForest):
    """
    Same as sklearn `RandomTreesEmbedding` but with distributed
    training using spark.

    Args:
        sc (sparkContext): Spark context for spark broadcasting and rdd operations.
        partitions (int or 'auto'): default 'auto'
            Number of partitions to use for parallelization of parameter
            search space. Integer values or None will be used directly for `numSlices`,
            while 'auto' will set `numSlices` to the number required fits.
        **kwargs: Keyword arguments to be passed to `RandomTreesEmbedding`.
    """
    criterion = 'mse'
    max_features = 1

    def __init__(self,
                 sc=None,
                 partitions='auto',
                 n_estimators=100,
                 max_depth=5,
                 min_samples_split=2,
                 min_samples_leaf=1,
                 min_weight_fraction_leaf=0.,
                 max_leaf_nodes=None,
                 min_impurity_decrease=0.,
                 min_impurity_split=None,
                 sparse_output=True,
                 n_jobs=None,
                 random_state=None,
                 verbose=0,
                 warm_start=False):
        super().__init__(
            base_estimator=ExtraTreeRegressor(),
            n_estimators=n_estimators,
            estimator_params=("criterion", "max_depth", "min_samples_split",
                              "min_samples_leaf", "min_weight_fraction_leaf",
                              "max_features", "max_leaf_nodes",
                              "min_impurity_decrease", "min_impurity_split",
                              "random_state"),
            bootstrap=False,
            oob_score=False,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            warm_start=warm_start)

        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease
        self.min_impurity_split = min_impurity_split
        self.sparse_output = sparse_output
        self.sc = sc
        self.partitions = partitions

    def _set_oob_score(self, X, y):
        raise NotImplementedError("OOB score not supported by tree embedding")

    def fit(self, X, y=None, sample_weight=None):
        """ Fit estimator """
        self.fit_transform(X, y, sample_weight=sample_weight)
        return self

    def fit_transform(self, X, y=None, sample_weight=None):
        """ Fit estimator and transform dataset """
        X = check_array(X, accept_sparse=['csc'])
        if issparse(X):
            # Pre-sort indices to avoid that each individual tree of the
            # ensemble sorts the indices.
            X.sort_indices()

        rnd = check_random_state(self.random_state)
        y = rnd.uniform(size=X.shape[0])
        super().fit(X, y, sample_weight=sample_weight)

        self.one_hot_encoder_ = OneHotEncoder(
            sparse=self.sparse_output, categories='auto')
        return self.one_hot_encoder_.fit_transform(self.apply(X))

    def transform(self, X):
        """ Transform dataset """
        check_is_fitted(self, "one_hot_encoder_")
        return self.one_hot_encoder_.transform(self.apply(X))



"""
Test ensemble classes
"""

try:
    import numpy as np
    from skdist.distribute import ensemble
    from skdist.distribute.ensemble import (
        DistRandomForestClassifier, DistRandomForestRegressor, 
        DistExtraTreesClassifier, DistExtraTreesRegressor,
        DistRandomTreesEmbedding
        )
    _import_error = None
except Exception as e:
    _import_error = e

def test_ensemble():
    assert _import_error == None

def test_rfc():
    X = np.array([[0,1,0,1], [0,0,0,1], [1,0,1,0]])
    y = np.array([0,1,0])
    rfc = ensemble.DistRandomForestClassifier(
        n_estimators=10, random_state=5)
    rfc.fit(X,y)
    preds = rfc.predict(X)
    assert np.allclose(preds, np.array([0,1,0]))
 
def test_rfr():
    X = np.array([[0,1,0,1], [0,0,0,1], [1,0,1,0]])
    y_reg = np.array([0.1, 0.2, 0.1])
    rfr = ensemble.DistRandomForestRegressor(
        n_estimators=10, random_state=5)
    rfr.fit(X,y_reg)
    preds = rfr.predict(X)
    assert np.allclose(preds, np.array([0.15, 0.18, 0.12]))

def test_etc():
    X = np.array([[0,1,0,1], [0,0,0,1], [1,0,1,0]])
    y = np.array([0,1,0])
    etc = ensemble.DistExtraTreesClassifier(
        n_estimators=10, random_state=5)
    etc.fit(X,y)
    preds = etc.predict(X)
    assert np.allclose(preds, np.array([0,1,0]))

def test_etr():
    X = np.array([[0,1,0,1], [0,0,0,1], [1,0,1,0]])
    y_reg = np.array([0.1, 0.2, 0.1])
    etr = ensemble.DistExtraTreesRegressor(
        n_estimators=10, random_state=5)  
    etr.fit(X,y_reg)
    preds = etr.predict(X)
    assert np.allclose(preds, np.array([0.1, 0.2, 0.1]))

def test_rte():
    X = np.array([[0,1,0,1], [0,0,0,1], [1,0,1,0]])
    rte = ensemble.DistRandomTreesEmbedding(
        n_estimators=10, random_state=5)  
    rte.fit(X,y=None)
    preds = rte.transform(X)
    assert preds.shape == (3,30)




"""
Various preprocessing classes implemented as scikit-learn
transformers compatible with scikit-learn pipelines.
"""

import warnings
import pandas as pd
import numpy as np

from scipy import sparse
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import FunctionTransformer, LabelEncoder, normalize
from sklearn.preprocessing.label import MultiLabelBinarizer
from sklearn.feature_extraction.text import HashingVectorizer
from sklearn import feature_selection

__all__ = [
    "SelectField", 
    "FeatureCast", 
    "ImputeNull", 
    "DenseTransformer", 
    "SparseTransformer", 
    "LabelEncoderPipe", 
    "SelectorMem", 
    "HashingVectorizerChunked",
    "MultihotEncoder"
    ]

_SELECTOR_LOOKUP = {
    "fpr": feature_selection.SelectFpr,
    "fdr": feature_selection.SelectFdr,
    "kbest": feature_selection.SelectKBest,
    "percentile": feature_selection.SelectPercentile,
    "fwe": feature_selection.SelectFwe
}

class _CustomFunctionTransformer(TransformerMixin, BaseEstimator):
    """ Base class for function transformers """
    def fit(self, X, y=None):
        self.transformer_.fit(X, y=None)
        return self

    def transform(self, X, y=None):
        return self.transformer_.transform(X)

def select_field(X, cols=None, single_dimension=False):
    """ 
    Select columns from a pandas DataFrame

    Args:
        X (pandas DataFrame): input data
        cols (array-like): list of columns to select
        single_dimension (bool): reduce data to 
            one dimension if only one column
            is requested
    Returns:
        X (numpy array)
    """
    if cols is None:
        return X.values
    if len(cols) > 1:
        return X[cols].values
    if len(cols) == 1:
        if single_dimension:
            return X[cols[0]].values
        else:
            return X[cols].values

class SelectField(_CustomFunctionTransformer):
    """ 
    Applies `select_field` as FunctionTransformer

    Args:
        cols (array-like): list of columns to select
        single_dimension (bool): reduce data to 
            one dimension if only one column
            is requested
    """
    def __init__(self, cols=None, single_dimension=False):
        self.cols = cols
        self.single_dimension = single_dimension
        kw_args = {"cols": cols, "single_dimension": single_dimension}
        self.transformer_ = FunctionTransformer(
            select_field, 
            kw_args=kw_args, 
            validate=False
            )

def to_dense(X):
    """ Densify sparse matrix if issparse """
    if sparse.issparse(X):
        return X.todense()
    else:
        return X

class DenseTransformer(_CustomFunctionTransformer):
    """ Applies `to_dense` as a FunctionTransformer """
    def __init__(self):
        self.transformer_ = FunctionTransformer(to_dense, validate=False)

def to_sparse(X):
    """ Sparsify dense matrix if not issparse """
    if sparse.issparse(X):
        return X
    else:
        return sparse.csr_matrix(X)

class SparseTransformer(_CustomFunctionTransformer):
    """ Applies `to_sparse` as a FunctionTransformer """
    def __init__(self):
        self.transformer_ = FunctionTransformer(to_sparse, validate=False)

def feature_cast(X, cast_type=None):
    """
    Casts feature data to requeted type

    Args:
        X (numpy array): input data
        cast_type (type): type to cast data
    Returns:
        X (numpy array)
    """
    if cast_type is None:
        return X
    else:
        return X.astype(cast_type)

class FeatureCast(_CustomFunctionTransformer):
    """
    Applies `feature_cast` as a FunctionTransformer

    Args:
        cast_type (type): type to cast data
    """
    def __init__(self, cast_type=None):
        self.transformer_ = FunctionTransformer(
            feature_cast, kw_args={"cast_type": cast_type}, 
            validate=False
            )

def impute_null(X, impute_val=None):
    """
    Impute null values. Null values are 
    determined using `pd.isnull`.

    Args:
        X (numpy array): input data
        impute_val (object): value to impute with
    Returns:
        X (numpy array)
    """
    if impute_val is None:
        return X
    else:
        X[pd.isnull(X)] = impute_val
        return X

class ImputeNull(_CustomFunctionTransformer):
    """
    Applies `impute_null` as a FunctionTransformer

    Args:
        impute_val (object): value to impute with
    """
    def __init__(self, impute_val=None):
        self.transformer_ = FunctionTransformer(
            impute_null, kw_args={"impute_val": impute_val},
            validate=False
            )

class LabelEncoderPipe(TransformerMixin, BaseEstimator):
    """ 
    LabelEncoder wrapper with TransformerMixin. Allows LabelEncoder 
    to work with a Pipeline. 
    """
    def fit(self, X, y=None):
        """ Fit the LabelEncoder """
        self.le = LabelEncoder()
        self.le.fit(X)
        return self
        
    def transform(self, X, y=None):
        """ Transform the label encoder """
        return np.array([[x] for x in self.le.transform(X)])

class SelectorMem(TransformerMixin, BaseEstimator):
    """ 
    Memory efficient feature selector. Identical functionality
    to classes in sklearn.feature_selection.univariate_selection but
    with more memory efficient attribute storage.
    
    Args:
        selector (object): any sklearn.feature_selection.univariate_selection._BaseFilter
            inheritor
        score_func (function): selector scoring function
        threshold (int/float): scoring threshold to use with scoring function and selector
    """
    def __init__(self, selector="fpr", score_func=feature_selection.f_classif, threshold=0.05):
        self.selector = selector
        self.score_func = score_func
        self.threshold = threshold

    def fit(self, X, y=None):
        """
        Run score function on (X, y) and get the appropriate features
        
        Args:
            X (array-like, shape = [n_samples, n_features])
                The training input samples
            y (array-like, shape = [n_samples])
                The target values (class labels in classification, real numbers in
                regression)
        """
        selector = _SELECTOR_LOOKUP[self.selector.lower()](self.score_func, self.threshold)
        selector.fit(X, y)
        mask_indices = selector.get_support(indices=True)
        mask_bool = selector.get_support(indices=False)
        if np.array(mask_bool).nbytes > np.array(mask_indices).nbytes:
            self.mask = mask_indices
        else:
            self.mask = mask_bool
        return self

    def transform(self, X, y=None):
        """
        Reduce X to the selected features

        Args:
            X (array of shape [n_samples, n_features])
                The input samples

        Returns:
            X_r (array of shape [n_samples, n_selected_features])
                The input samples with only the selected features
        """
        return X[:, self.mask]

class HashingVectorizerChunked(HashingVectorizer):
    """
    Equivalent to HashingVectorizer but with chunked prediction

    Args:
        chunksize (int): size of transform chunks
    """
    def __init__(self, chunksize=100000, **kwargs):
        self.chunksize = chunksize
        HashingVectorizer.__init__(
            self, **kwargs)

    def transform(self, X):
        """Transform a sequence of documents to a document-term matrix

        Args:
            X (iterable over raw text documents, length = n_samples)
                Samples. Each sample must be a text document (either bytes or
                unicode strings, file name or file object depending on the
                constructor argument) which will be tokenized and hashed.

        Returns:
            X (scipy.sparse matrix, shape = (n_samples, self.n_features))
                Document-term matrix.
        """
        if (len(X) < self.chunksize) or (self.chunksize is None):
            return self._transform(X)
        else:
            return sparse.vstack([self._transform(x) for x in self._chunks(X)])

    def _transform(self, X):
        if isinstance(X, str):
            raise ValueError(
                "Iterable over raw text documents expected, "
                "string object received.")

        analyzer = self.build_analyzer()
        X = self._get_hasher().transform(analyzer(doc) for doc in X)
        if self.binary:
            X.data.fill(1)
        if self.norm is not None:
            X = normalize(X, norm=self.norm, copy=False)
        return X

    def _chunks(self, l):
        for i in range(0, len(l), self.chunksize):
            yield l[i:i + self.chunksize]

class MultihotEncoder(BaseEstimator, TransformerMixin):
    """
    Wraps `MultiLabelBinarizer` in a pipeline safe transformer

    Args:
        sparse_output (bool): convert output to sparse matrix
    """
    def __init__(self, sparse_output=False):
        self.transformer = MultiLabelBinarizer()
        self.sparse_output = sparse_output

    def fit(self, X, y=None):
        """ Fit MultiLabelBinarizer """
        self.transformer.fit(X)
        return self

    def transform(self, X,y=None):
        """ Transform MultiLabelBinarizer """
        # ignore unseen label warning
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            X_t = self.transformer.transform(X)
        if self.sparse_output:
            return sparse.csr_matrix(X_t)
        else:
            return X_t



"""
Validation functions for the distribute module
"""

import numpy as np

from sklearn.model_selection import ParameterGrid

def _check_estimator(estimator, verbose=False):
    """ Print sparkContext awareness if apporpriate """
    if verbose:
        if estimator.sc is None:
            print("No spark context is provided; running locally")
        else:
            print("Spark context found; running with spark")

def _validate_params(param_sets):
    """ Validates grid params """
    invalid_params = [
        type(param) 
        for param in param_sets 
        if not isinstance(param, dict)
        ]
    if invalid_params:
        raise ValueError('Params must be dictionaries: got '
            '{0!r}'.format(invalid_params))  

def _validate_models(models, clf):
    """ Validates models input argument """
    try:
        iter(models)
    except:
        raise TypeError("Input argument 'models' is not iterable.")
    if not (isinstance(models[0], tuple) or isinstance(models[0], list)):
        models = [models]
    names, estimators, param_sets = zip(*models)
    _validate_names(clf, names)
    _validate_estimators(estimators)
    _validate_params(param_sets)
    return models

def _check_base_estimator(estimator):
    """ Validates base estimator """
    return (
        hasattr(estimator, "fit") and 
        hasattr(estimator, "predict")
        )
        
def _validate_estimators(estimators):
    """ Validates estimators """
    invalid_estimators = [
        type(estimator) 
        for estimator in estimators 
        if not _check_base_estimator(estimator)
        ]
    if invalid_estimators:
        raise ValueError('Estimators must be sklearn estimators: got '
            '{0!r}'.format(invalid_estimators))

def _validate_names(cls, names):
    """ Validates names """
    if len(set(names)) != len(names):
        raise ValueError('Names provided are not unique: '
            '{0!r}'.format(list(names)))
    invalid_names = [name for name in names if not isinstance(name, str)]
    if invalid_names:
        raise ValueError('Estimator names must be strings: got '
            '{0!r}'.format(invalid_names))
    invalid_names = set(names).intersection(cls.get_params(deep=False))
    if invalid_names:
        raise ValueError('Estimator names conflict with constructor '
            'arguments: {0!r}'.format(sorted(invalid_names)))
    invalid_names = [name for name in names if '__' in name]
    if invalid_names:
        raise ValueError('Estimator names must not contain __: got '
            '{0!r}'.format(invalid_names))

def _check_n_iter(n_iter, param_distributions):
    """
    Check if n_iter is greater than the total number 
    of possible param sets from the given distribution.
    """
    all_lists = np.all([not hasattr(v, "rvs") 
        for v in param_distributions.values()])
    if all_lists:
        param_grid = ParameterGrid(param_distributions)
        grid_size = len(param_grid)
    else:
        grid_size = n_iter
    return min(grid_size, n_iter)



"""
Test preprocessing
"""

try:
    import pandas as pd
    import numpy as np
    from scipy.sparse import csr_matrix
    from skdist import preprocessing
    _import_error = None
except Exception as e:
    _import_error = e

def test_import_preprocessing():
    assert _import_error == None

def test_varselect():
    vars = ["text"]
    df = pd.DataFrame({"text": ["a", "b", "c"], "numbers": [1, 2, 3]})
    selector = preprocessing.SelectField(cols=vars, single_dimension=True)
    X_t = selector.fit_transform(df)
    assert X_t.shape == (3,)

def test_varselect_multidim():
    vars = ["text"]
    df = pd.DataFrame({"text": ["a", "b", "c"], "numbers": [1, 2, 3]})
    selector = preprocessing.SelectField(cols=vars, single_dimension=False)
    X_t = selector.fit_transform(df)
    assert X_t.shape == (3,1)

def test_featurecast():
    data = np.array([[1,2,3]])
    X_t = preprocessing.FeatureCast(cast_type=float).fit_transform(data)
    assert isinstance(X_t[0][0], np.float64)

def test_fillna():
    X = np.array([["a", np.nan, "c"], [np.nan, 2, 3]])
    selector = preprocessing.ImputeNull("a")
    X_t = selector.fit_transform(X)
    assert X_t.shape == X.shape

def test_densetransformer():
    data = np.array([[1,2,3], [4,5,6]])
    X = csr_matrix(data)
    X_t = preprocessing.DenseTransformer().fit_transform(X)
    assert isinstance(X_t, np.ndarray)

def test_labelencoderpipe():
    X = np.array([1,2,3])
    X_t = preprocessing.LabelEncoderPipe().fit_transform(X)
    assert X_t.shape[0] == X.shape[0]

def test_le_pipe():
    X = np.array([1,2,3])
    X_t = preprocessing.LabelEncoderPipe().fit_transform(X)
    assert X_t.shape != X.shape

def test_selectormem():
    X = np.array([[1,2,3], [4,5,6], [10,10,10]])
    y = np.array([0,0,1])

    X_t = preprocessing.SelectorMem().fit_transform(X,y)
    assert X_t.shape != X.shape

def test_hashingvectorizerchunked():
    X = ["here is some text", "more text here", "this is also text"]
    X_t = preprocessing.HashingVectorizerChunked().fit_transform(X)
    assert X_t.shape == (3,2**20)

def test_hv_chunked():
    X = ["here is some text", "more text here", "this is also text"]
    X_t = preprocessing.HashingVectorizerChunked(chunksize=1).fit_transform(X)
    assert X_t.shape == (3,2**20)

def test_multihot():
    X = [["text", "more text", "this"], ["more text"], ["new text"]]
    X_t = preprocessing.MultihotEncoder().fit_transform(X)
    assert X_t.shape == (3,4)

def test_multihot_sparse():
    X = [["text", "more text", "this"], ["more text"], ["new text"]]
    X_t = preprocessing.MultihotEncoder(sparse_output=True).fit_transform(X)
    assert X_t.shape == (3,4)
  
def test_multihot_unknown():
    X = [["text", "more text", "this"], ["more text"], ["new text"]]
    enc = preprocessing.MultihotEncoder()
    enc.fit(X)
    X_t = enc.transform([["unseen", "this"], ["more text"], ["unknown"]])
    assert X_t.shape == (3,4)
    



"""
Utility functions for the distribute module
"""

import numbers
import numpy as np

from sklearn.utils import safe_indexing
from sklearn.metrics.scorer import check_scoring

def _aggregate_score_dicts(scores):
    """ Aggregate the list of dict to dict of np ndarray """
    return {key: np.asarray([score[key] for score in scores])
            for key in scores[0]}

def _multimetric_score(estimator, X_test, y_test, scorers):
    """ Return a dict of score for multimetric scoring """
    scores = {}

    for name, scorer in scorers.items():
        if y_test is None:
            score = scorer(estimator, X_test)
        else:
            score = scorer(estimator, X_test, y_test)

        if hasattr(score, 'item'):
            try:
                # e.g. unwrap memmapped scalars
                score = score.item()
            except ValueError:
                # non-scalar?
                pass
        scores[name] = score

        if not isinstance(score, numbers.Number):
            raise ValueError("scoring must return a number, got %s (%s) "
                             "instead. (scorer=%s)"
                             % (str(score), type(score), name))
    return scores

def _score(estimator, X_test, y_test, scorer, is_multimetric=False):
    """
    Compute the score(s) of an estimator on a given test set. Will return
    a single float if is_multimetric is False and a dict of floats,
    if is_multimetric is True
    """
    if is_multimetric:
        return _multimetric_score(estimator, X_test, y_test, scorer)
    else:
        if y_test is None:
            score = scorer(estimator, X_test)
        else:
            score = scorer(estimator, X_test, y_test)

        if hasattr(score, 'item'):
            try:
                # e.g. unwrap memmapped scalars
                score = score.item()
            except ValueError:
                # non-scalar?
                pass

        if not isinstance(score, numbers.Number):
            raise ValueError("scoring must return a number, got %s (%s) "
                             "instead. (scorer=%r)"
                             % (str(score), type(score), scorer))
    return score

def _check_multimetric_scoring(estimator, scoring=None):
    """ Check the scoring parameter in cases when multiple metrics are allowed """
    if callable(scoring) or scoring is None or isinstance(scoring,
                                                          str):
        scorers = {"score": check_scoring(estimator, scoring=scoring)}
        return scorers, False
    else:
        err_msg_generic = ("scoring should either be a single string or "
                           "callable for single metric evaluation or a "
                           "list/tuple of strings or a dict of scorer name "
                           "mapped to the callable for multiple metric "
                           "evaluation. Got %s of type %s"
                           % (repr(scoring), type(scoring)))

        if isinstance(scoring, (list, tuple, set)):
            err_msg = ("The list/tuple elements must be unique "
                       "strings of predefined scorers. ")
            invalid = False
            try:
                keys = set(scoring)
            except TypeError:
                invalid = True
            if invalid:
                raise ValueError(err_msg)

            if len(keys) != len(scoring):
                raise ValueError(err_msg + "Duplicate elements were found in"
                                 " the given list. %r" % repr(scoring))
            elif len(keys) > 0:
                if not all(isinstance(k, str) for k in keys):
                    if any(callable(k) for k in keys):
                        raise ValueError(err_msg +
                                         "One or more of the elements were "
                                         "callables. Use a dict of score name "
                                         "mapped to the scorer callable. "
                                         "Got %r" % repr(scoring))
                    else:
                        raise ValueError(err_msg +
                                         "Non-string types were found in "
                                         "the given list. Got %r"
                                         % repr(scoring))
                scorers = {scorer: check_scoring(estimator, scoring=scorer)
                           for scorer in scoring}
            else:
                raise ValueError(err_msg +
                                 "Empty list was given. %r" % repr(scoring))

        elif isinstance(scoring, dict):
            keys = set(scoring)
            if not all(isinstance(k, str) for k in keys):
                raise ValueError("Non-string types were found in the keys of "
                                 "the given dict. scoring=%r" % repr(scoring))
            if len(keys) == 0:
                raise ValueError("An empty dict was passed. %r"
                                 % repr(scoring))
            scorers = {key: check_scoring(estimator, scoring=scorer)
                       for key, scorer in scoring.items()}
        else:
            raise ValueError(err_msg_generic)
        return scorers, True

def _num_samples(x):
    """ Return number of samples in array-like x """
    if hasattr(x, 'fit') and callable(x.fit):
        # Don't get num_samples from an ensembles length!
        raise TypeError('Expected sequence or array-like, got '
                        'estimator %s' % x)
    if not hasattr(x, '__len__') and not hasattr(x, 'shape'):
        if hasattr(x, '__array__'):
            x = np.asarray(x)
        else:
            raise TypeError("Expected sequence or array-like, got %s" %
                            type(x))
    if hasattr(x, 'shape'):
        if len(x.shape) == 0:
            raise TypeError("Singleton array %r cannot be considered"
                            " a valid collection." % x)
        # Check that shape is returning an integer or default to len
        # Dask dataframes may not return numeric shape[0] value
        if isinstance(x.shape[0], numbers.Integral):
            return x.shape[0]
        else:
            return len(x)
    else:
        return len(x)

def _safe_split(estimator, X, y, indices, train_indices=None):
    """Create subset of dataset and properly handle kernels"""
    from sklearn.gaussian_process.kernels import Kernel as GPKernel

    if (hasattr(estimator, 'kernel') and callable(estimator.kernel) and
            not isinstance(estimator.kernel, GPKernel)):
        # cannot compute the kernel values with custom function
        raise ValueError("Cannot use a custom kernel function. "
                         "Precompute the kernel matrix instead.")

    if not hasattr(X, "shape"):
        if getattr(estimator, "_pairwise", False):
            raise ValueError("Precomputed kernels or affinity matrices have "
                             "to be passed as arrays or sparse matrices.")
        X_subset = [X[index] for index in indices]
    else:
        if getattr(estimator, "_pairwise", False):
            # X is a precomputed square kernel matrix
            if X.shape[0] != X.shape[1]:
                raise ValueError("X should be a square kernel matrix")
            if train_indices is None:
                X_subset = X[np.ix_(indices, indices)]
            else:
                X_subset = X[np.ix_(indices, train_indices)]
        else:
            X_subset = safe_indexing(X, indices)

    if y is not None:
        y_subset = safe_indexing(y, indices)
    else:
        y_subset = None
    return X_subset, y_subset

def _dict_slice_remove(a, b, cols=["model_index", "param_set"]):
    """ 
    Remove dictionary records from list 'a' 
    if the records appear in 'b' 
    for the given columns 'cols'.
    """
    b_ = [{k:d[k] for k in cols} for d in b]
    a_ = []
    for d in a:
        if {k:d[k] for k in cols} not in b_:
            a_.append(d)
    return a_




"""
Test postprocessing 
"""

try:
    import numpy as np
    from sklearn.linear_model import LogisticRegression
    from skdist import postprocessing
    _import_error = None
except Exception as e:
    _import_error = e

def test_import_postprocessing():
    assert _import_error == None

def test_predict():
    X = np.array([[1,2,3], [4,5,6]])
    y = np.array([0,1])

    model1 = LogisticRegression(solver="liblinear")
    model2 = LogisticRegression(solver="lbfgs")
    model1.fit(X,y)
    model2.fit(X,y)
    
    clf = postprocessing.SimpleVoter(
        [("model1", model1), ("model2", model2)],
        voting="soft", classes=model1.classes_
        )
    pred = clf.predict(X)
    probs = clf.predict_proba(X)
    assert pred.shape == y.shape

def test_predict_hard_voting():
    X = np.array([[1,2,3], [4,5,6]])
    y = np.array([0,1])

    model1 = LogisticRegression(solver="liblinear")
    model2 = LogisticRegression(solver="lbfgs")
    model1.fit(X,y)
    model2.fit(X,y)

    clf = postprocessing.SimpleVoter(
        [("model1", model1), ("model2", model2)],
        voting="hard", classes=model1.classes_
        )
    pred = clf.predict(X)
    assert pred.shape == y.shape



"""
Distributed grid search meta-estimators
"""

import time
import numbers
import random
import numpy as np
import pandas as pd

from copy import copy
from abc import ABCMeta
from joblib import Parallel, delayed
from sklearn.model_selection import (
    ParameterGrid, GridSearchCV, 
    RandomizedSearchCV, ParameterSampler, 
    check_cv
    )
from sklearn.metrics.scorer import check_scoring
from sklearn.base import BaseEstimator, is_classifier
from sklearn.utils.metaestimators import if_delegate_has_method
from sklearn.utils.validation import indexable, check_is_fitted
from sklearn.utils.fixes import MaskedArray

from functools import partial
from scipy.stats import rankdata
from itertools import product
from collections import defaultdict

from .validation import (
    _check_estimator, _check_base_estimator, 
    _validate_params, _validate_models, 
    _validate_names, _validate_estimators, 
    _check_n_iter
    )
from .utils import (
    _multimetric_score, _num_samples, 
    _aggregate_score_dicts, _score, 
    _check_multimetric_scoring,
    _safe_split, _dict_slice_remove
    )
from .base import (
    _clone, _get_value, _parse_partitions
    )

__all__ = [
    "DistGridSearchCV", 
    "DistRandomizedSearchCV",
    "DistMultiModelSearch"
]

def _sample_one(n_iter, param_distributions, random_state=None):
    """ Sample from param distributions for one model """
    return list(ParameterSampler(
        param_distributions, 
        n_iter=_check_n_iter(n_iter, param_distributions), 
        random_state=random_state
        ))

def _raw_sampler(models, n_params=None, n=None, 
                 random_state=None):
    """ Sample from param distributions for each model """
    if n_params is None:
        if n is None:
            raise Exception(
                "Must supply either 'n_params' or 'n' as arguments")
        else:
            n_params = [n]*len(models)
    param_sets = []
    for index in range(len(models)):
        sampler = _sample_one(
            n_params[index], models[index][2], 
            random_state=random_state
            )
        for sample_index in range(len(sampler)):
            param_set = {
                "model_index": index, 
                "params_index": sample_index, 
                "param_set": sampler[sample_index]
                }
            param_sets.append(param_set)
    return param_sets

def _fit_one_fold(fit_set, models, X, y, scoring, fit_params={}):
    """
    Fits the given estimator on one fold of training data.
    Scores the fitted estimator against the test fold.
    """
    train = fit_set[0][0]
    test = fit_set[0][1]
    estimator_ = _clone(models[fit_set[1]["model_index"]][1])
    parameters = fit_set[1]["param_set"]
    X_train, y_train = _safe_split(estimator_, X, y, train)
    X_test, y_test = _safe_split(estimator_, X, y, test, train)
    if parameters is not None:
        estimator_.set_params(**parameters)
    estimator_.fit(X_train, y_train, **fit_params)
    scorer = check_scoring(estimator_, scoring=scoring)
    is_multimetric = not callable(scorer)
    out_dct = fit_set[1]
    out_dct["score"] = _score(
        estimator_, X_test, y_test, 
        scorer, is_multimetric
        )
    return out_dct

def _fit_batch(X, y, folds, param_sets, models, n, 
               scoring, random_state=None, sc=None, 
               partitions="auto", n_jobs=None, fit_params={}):
    """
    Fits a batch of combinations of parameter sets, models
    and cross validation folds. Returns results pandas
    DataFrames.
    """
    fit_sets = product(folds, param_sets)
    if sc is None:
        scores = Parallel(n_jobs=n_jobs)(
            delayed(_fit_one_fold)(
                x, models, X, y, 
                scoring, fit_params
                )
            for x in fit_sets)
    else:
        fit_sets = list(fit_sets)
        partitions = _parse_partitions(partitions, len(fit_sets))
        scores = (
            sc
            .parallelize(fit_sets, numSlices=partitions)
            .map(lambda x: _fit_one_fold(x, models, X, y, scoring, fit_params))
            .collect()
            )
    param_results = _get_results(scores)
    model_results = (
        param_results
        .groupby(["model_index"])["score"]
        .max()
        .reset_index()
        .sort_values("model_index")
        )
    return param_results, model_results

def _get_results(scores):
    """ Converts 'scores' list to pandas DataFrame """
    cols = [
        "model_index", "params_index", 
        "param_set", "score"
        ]
    df = (
        pd.DataFrame(scores, columns=cols)
        .sort_values(["model_index", "params_index"])
        )
    if len(df) == 0:
        return pd.DataFrame(columns=cols)
    return (
        df
        .groupby(["model_index", "params_index"])
        .agg({"score": "mean", "param_set": "first"})
        .reset_index()
        .sort_values(["model_index", "params_index"])
        [cols]
        )

def _fit_and_score(estimator, X, y, scorer, train, test, verbose,
                   parameters, fit_params, return_train_score=False,
                   return_parameters=False, return_n_test_samples=False,
                   return_times=False, error_score='raise'):
    """ Fit estimator and compute scores for a given dataset split """
    estimator_ = _clone(_get_value(estimator))
    if verbose > 1:
        if parameters is None:
            msg = ''
        else:
            msg = '%s' % (', '.join('%s=%s' % (k, v)
                          for k, v in list(parameters.items())))
        print(("[CV] %s %s" % (msg, (64 - len(msg)) * '.')))

    # Adjust length of sample weights
    fit_params = fit_params if fit_params is not None else {}
    fit_params = dict([(k, _index_param_value(X, v, train))
                      for k, v in list(fit_params.items())])

    test_scores = {}
    train_scores = {}
    if parameters is not None:
        estimator_.set_params(**parameters)

    start_time = time.time()

    X_train, y_train = _safe_split(estimator_, X, y, train)
    X_test, y_test = _safe_split(estimator_, X, y, test, train)

    is_multimetric = not callable(scorer)
    n_scorers = len(list(scorer.keys())) if is_multimetric else 1

    try:
        if y_train is None:
            estimator_.fit(X_train, **fit_params)
        else:
            estimator_.fit(X_train, y_train, **fit_params)

    except Exception as e:
        # Note fit time as time until error
        fit_time = time.time() - start_time
        score_time = 0.0
        if error_score == 'raise':
            raise
        elif isinstance(error_score, numbers.Number):
            if is_multimetric:
                test_scores = dict(list(zip(list(scorer.keys()),
                                   [error_score, ] * n_scorers)))
                if return_train_score:
                    train_scores = dict(list(zip(list(scorer.keys()),
                                        [error_score, ] * n_scorers)))
            else:
                test_scores = error_score
                if return_train_score:
                    train_scores = error_score
            warnings.warn("Classifier fit failed. The score on this train-test"
                          " partition for these parameters will be set to %f. "
                          "Details: \n%r" % (error_score, e), FitFailedWarning)
        else:
            raise ValueError("error_score must be the string 'raise' or a"
                             " numeric value. (Hint: if using 'raise', please"
                             " make sure that it has been spelled correctly.)")

    else:
        fit_time = time.time() - start_time
        # _score will return dict if is_multimetric is True
        test_scores = _score(estimator_, X_test, y_test, scorer, is_multimetric)
        score_time = time.time() - start_time - fit_time
        if return_train_score:
            train_scores = _score(estimator_, X_train, y_train, scorer,
                                  is_multimetric)

    if verbose > 2:
        if is_multimetric:
            for scorer_name, score in list(test_scores.items()):
                msg += ", %s=%s" % (scorer_name, score)
        else:
            msg += ", score=%s" % test_scores
    if verbose > 1:
        total_time = score_time + fit_time
        end_msg = "%s, total=%s" % (msg, logger.short_format_time(total_time))
        print(("[CV] %s %s" % ((64 - len(end_msg)) * '.', end_msg)))

    ret = [train_scores, test_scores] if return_train_score else [test_scores]

    if return_n_test_samples:
        ret.append(_num_samples(X_test))
    if return_times:
        ret.extend([fit_time, score_time])
    if return_parameters:
        ret.append(parameters)
    return ret
    
class DistBaseSearchCV(BaseEstimator, metaclass=ABCMeta):
    """
    Same as sklearn `BaseSearchCV` but with distributed
    training using spark

    Args:
        estimator (estimator object):
            This is assumed to implement the scikit-learn estimator interface.
            Either estimator needs to provide a ``score`` function,
            or ``scoring`` must be passed.
        sc (sparkContext): Spark context for spark broadcasting and rdd operations.
        partitions (int or 'auto'): default 'auto'
            Number of partitions to use for parallelization of parameter
            search space. Integer values or None will be used directly for `numSlices`,
            while 'auto' will set `numSlices` to the number required fits.
        preds (bool): keep predictions as attribute
    """
    def __init__(self, estimator, sc=None, partitions='auto', preds=False):
        self.estimator = estimator
        self.sc = sc
        self.partitions = partitions
        self.preds = preds
        
    def fit(self, X, y=None, groups=None, **fit_params):
        """
        Run fit with all sets of parameters. Parallelize fit operations
        using spark.
        
        Args:
            X (array-like, shape = [n_samples, n_features]): training vector, 
                where n_samples is the number of samples and
                n_features is the number of features
            y (array-like, shape = [n_samples] or [n_samples, n_output]): target 
                relative to X for classification or regression
            groups (array-like, with shape (n_samples,)): group labels for 
                the samples used while splitting the dataset into
                train/test set
            **fit_params (dict of string -> object): parameters passed 
                to the ``fit`` method of the estimator
        """
        _check_estimator(self, verbose=self.verbose)
        estimator = self.estimator
        cv = check_cv(self.cv, y, classifier=is_classifier(estimator))

        scorers, self.multimetric_ = _check_multimetric_scoring(
            self.estimator, scoring=self.scoring)

        if self.multimetric_:
            if self.refit is not False and (
                    not isinstance(self.refit, six.string_types) or
                    # This will work for both dict / list (tuple)
                    self.refit not in scorers):
                raise ValueError("For multi-metric scoring, the parameter "
                                 "refit must be set to a scorer key "
                                 "to refit an estimator with the best "
                                 "parameter setting on the whole data and "
                                 "make the best_* attributes "
                                 "available for that metric. If this is not "
                                 "needed, refit should be set to False "
                                 "explicitly. %r was passed." % self.refit)
            else:
                refit_metric = self.refit
        else:
            refit_metric = 'score'

        X, y, groups = indexable(X, y, groups)
        n_splits = cv.get_n_splits(X, y, groups)
        # Regenerate parameter iterable for each fit
        candidate_params = list(self._get_param_iterator())
        n_candidates = len(candidate_params)
        if self.verbose > 0:
            print(("Fitting {0} folds for each of {1} candidates, totalling"
                  " {2} fits".format(n_splits, n_candidates,
                                     n_candidates * n_splits)))

        base_estimator = _clone(self.estimator)
        pre_dispatch = self.pre_dispatch

        fit_sets = []
        cv_splitted = list(cv.split(X, y, groups))
        count = -1
        for fit_set in product(candidate_params, cv_splitted):
            count += 1
            fit_sets.append((count,) + fit_set)
        verbose = self.verbose
        return_train_score = self.return_train_score
        error_score = self.error_score

        if self.sc is None:
            base_estimator_ = base_estimator
            out = Parallel(n_jobs=self.n_jobs)(
                delayed(_fit_and_score)(
                    base_estimator_, X, y, scorers, x[2][0], x[2][1],
                    verbose, x[1], fit_params=fit_params,
                    return_train_score=return_train_score,
                    return_n_test_samples=True,
                    return_times=True, return_parameters=False,
                    error_score=error_score)
                for x in fit_sets)
            out = [[fit_sets[ind][0], out[ind]] for ind in range(len(fit_sets))]
        else:
            base_estimator_ = self.sc.broadcast(base_estimator)
            partitions = _parse_partitions(self.partitions, len(fit_sets))
            out = self.sc.parallelize(fit_sets, numSlices=partitions).map(lambda x: [x[0], _fit_and_score(
                base_estimator_, X, y, scorers, x[2][0], x[2][1], 
                verbose, x[1], fit_params=fit_params,
                return_train_score=return_train_score,
                return_n_test_samples=True,
                return_times=True, return_parameters=False,
                error_score=error_score)]).collect()

        out = [out[i][1] for i in np.argsort([x[0] for x in out])]

        # if one choose to see train score, "out" will contain train score info
        if self.return_train_score:
            (train_score_dicts, test_score_dicts, test_sample_counts, fit_time,
             score_time) = list(zip(*out))
        else:
            (test_score_dicts, test_sample_counts, fit_time,
             score_time) = list(zip(*out))

        # test_score_dicts and train_score dicts are lists of dictionaries and
        # we make them into dict of lists
        test_scores = _aggregate_score_dicts(test_score_dicts)
        if self.return_train_score:
            train_scores = _aggregate_score_dicts(train_score_dicts)

        results = {}

        def _store(key_name, array, weights=None, splits=False, rank=False):
            """ A small helper to store the scores/times to the cv_results_ """
            array = np.array(array, dtype=np.float64).reshape(n_candidates,
                                                              n_splits)
            if splits:
                for split_i in range(n_splits):
                    # Uses closure to alter the results
                    results["split%d_%s"
                            % (split_i, key_name)] = array[:, split_i]

            array_means = np.average(array, axis=1, weights=weights)
            results['mean_%s' % key_name] = array_means
            # Weighted std is not directly available in numpy
            array_stds = np.sqrt(np.average((array -
                                             array_means[:, np.newaxis]) ** 2,
                                            axis=1, weights=weights))
            results['std_%s' % key_name] = array_stds

            if rank:
                results["rank_%s" % key_name] = np.asarray(
                    rankdata(-array_means, method='min'), dtype=np.int32)

        _store('fit_time', fit_time)
        _store('score_time', score_time)
        param_results = defaultdict(partial(MaskedArray,
                                            np.empty(n_candidates,),
                                            mask=True,
                                            dtype=object))
        for cand_i, params in enumerate(candidate_params):
            for name, value in list(params.items()):
                # An all masked empty array gets created for the key
                # `"param_%s" % name` at the first occurence of `name`.
                # Setting the value at an index also unmasks that index
                param_results["param_%s" % name][cand_i] = value

        results.update(param_results)
        # Store a list of param dicts at the key 'params'
        results['params'] = candidate_params

        # NOTE test_sample counts (weights) remain the same for all candidates
        test_sample_counts = np.array(test_sample_counts[:n_splits],
                                      dtype=np.int)
        for scorer_name in list(scorers.keys()):
            # Computed the (weighted) mean and std for test scores alone
            _store('test_%s' % scorer_name, test_scores[scorer_name],
                   splits=True, rank=True,
                   weights=test_sample_counts if self.iid else None)
            if self.return_train_score:
                prev_keys = set(results.keys())
                _store('train_%s' % scorer_name, train_scores[scorer_name],
                       splits=True)

                if self.return_train_score == 'warn':
                    for key in set(results.keys()) - prev_keys:
                        message = (
                            'You are accessing a training score ({!r}), '
                            'which will not be available by default '
                            'any more in 0.21. If you need training scores, '
                            'please set return_train_score=True').format(key)
                        # warn on key access
                        results.add_warning(key, message, FutureWarning)

        # For multi-metric evaluation, store the best_index_, best_params_ and
        # best_score_ iff refit is one of the scorer names
        # In single metric evaluation, refit_metric is "score"
        if self.refit or not self.multimetric_:
            self.best_index_ = results["rank_test_%s" % refit_metric].argmin()
            self.best_params_ = candidate_params[self.best_index_]
            self.best_score_ = results["mean_test_%s" % refit_metric][
                self.best_index_]

        if self.refit:
            self.best_estimator_ = _clone(base_estimator).set_params(
                **self.best_params_)
            if y is not None:
                self.best_estimator_.fit(X, y, **fit_params)
            else:
                self.best_estimator_.fit(X, **fit_params)
            if self.preds:
                preds = []
                for train_index, test_index in cv_splitted:
                    estimator_ = _clone(base_estimator).set_params(
                        **self.best_params_)
                    estimator_.fit(X[train_index], y[train_index])
                    try:
                        preds.append(estimator_.predict_proba(X[test_index]))
                    except:
                        preds.append(estimator_.predict(X[test_index]))
                self.preds_ = np.vstack(preds)

        # Store the only scorer not as a dict for single metric evaluation
        self.scorer_ = scorers if self.multimetric_ else scorers['score']

        self.cv_results_ = results
        self.n_splits_ = n_splits

        del self.sc
        if hasattr(self.estimator, "sc"):
            del self.estimator.sc
        return self

    def get_preds(self):
        """ Get CV predictions """
        if hasattr(self, "preds_"):
            return self.preds_

    def drop_preds(self):
        """ Remove preds_ attribute """
        if hasattr(self, "preds_"):
            del self.preds_
        
class DistGridSearchCV(DistBaseSearchCV, GridSearchCV):
    """
    Same as sklearn `GridSearchCV` but with distributed
    training using spark.

    Args:
        estimator (estimator object):
            This is assumed to implement the scikit-learn estimator interface.
            Either estimator needs to provide a ``score`` function,
            or ``scoring`` must be passed.
        param_grid (dict or list of dictionaries):
            Dictionary with parameters names (string) as keys and lists of
            parameter settings to try as values, or a list of such
            dictionaries, in which case the grids spanned by each dictionary
            in the list are explored. This enables searching over any sequence
            of parameter settings.
        sc (sparkContext): Spark context for spark broadcasting and rdd operations.
        partitions (int or 'auto'): default 'auto'
            Number of partitions to use for parallelization of parameter
            search space. Integer values or None will be used directly for `numSlices`,
            while 'auto' will set `numSlices` to the number required fits.
        preds (bool): keep predictions as attribute
    """
    def __init__(self, 
                 estimator, 
                 param_grid, 
                 sc=None, 
                 partitions='auto', 
                 preds=False, 
                 scoring=None,
                 n_jobs=None, 
                 iid='warn', 
                 refit=True, 
                 cv=5, 
                 verbose=0,
                 pre_dispatch='2*n_jobs', 
                 error_score='raise-deprecating',
                 return_train_score=False):
        GridSearchCV.__init__(
            self, 
            estimator, 
            param_grid, 
            scoring=scoring,
            n_jobs=n_jobs, 
            iid=iid, 
            refit=refit, 
            cv=cv, 
            verbose=verbose,
            pre_dispatch=pre_dispatch, 
            error_score=error_score,
            return_train_score=return_train_score
            )
        self.sc = sc
        self.partitions = partitions
        self.preds = preds
        
    def _get_param_iterator(self):
        """ Return ParameterGrid instance for the given param_grid """
        return ParameterGrid(self.param_grid)
        
class DistRandomizedSearchCV(DistBaseSearchCV, RandomizedSearchCV):
    """
    Same as sklearn `RandomizedSearchCV` but with distributed
    training using spark.

    Args:
        estimator (estimator object):
            This is assumed to implement the scikit-learn estimator interface.
            Either estimator needs to provide a ``score`` function,
            or ``scoring`` must be passed.
        param_distributions (dict):
            Dictionary with parameters names (string) as keys and distributions
            or lists of parameters to try. Distributions must provide a ``rvs``
            method for sampling (such as those from scipy.stats.distributions).
            If a list is given, it is sampled uniformly.
        sc (sparkContext): Spark context for spark broadcasting and rdd operations.
        partitions (int or 'auto'): default 'auto'
            Number of partitions to use for parallelization of parameter
            search space. Integer values or None will be used directly for `numSlices`,
            while 'auto' will set `numSlices` to the number required fits.
        preds (bool): keep predictions as attribute
    """
    def __init__(self, 
                 estimator, 
                 param_distributions, 
                 sc=None, 
                 partitions='auto', 
                 preds=False, 
                 n_iter=10, 
                 scoring=None,
                 n_jobs=None, 
                 iid='warn', 
                 refit=True,
                 cv=5, 
                 verbose=0, 
                 pre_dispatch='2*n_jobs',
                 random_state=None, 
                 error_score='raise-deprecating',
                 return_train_score=False):
        RandomizedSearchCV.__init__(
            self, 
            estimator, 
            param_distributions, 
            n_iter=n_iter, 
            scoring=scoring,
            n_jobs=n_jobs, 
            iid=iid, 
            refit=refit,
            cv=cv, 
            verbose=verbose, 
            pre_dispatch=pre_dispatch,
            random_state=random_state, 
            error_score=error_score,
            return_train_score=return_train_score
            )
        self.sc = sc
        self.partitions = partitions
        self.preds = preds
        
    def _get_param_iterator(self):
        """ Return ParameterSampler instance for the given distributions """
        return ParameterSampler(
            self.param_distributions, self.n_iter,
            random_state=self.random_state)

class DistMultiModelSearch(BaseEstimator, metaclass=ABCMeta):
    """
    Distributed multi-model search meta-estimator. Similar to 
    `DistRandomizedSearchCV` but with handling for multiple models.
    Takes a `models` input containing a list of tuples, each with a 
    string name, instantiated estimator object, and parameter set
    dictionary for random search.

    The fit method will compute a cross validation score for each
    estimator/parameter set combination after randomly sampling from
    the parameter set for each estimator. The best estimator/parameter
    set combination will be refit if appropriate. The process is distrubuted
    with spark if a sparkContext is provided, else joblib is used.

    Args:
        models (array-like): List of tuples containing estimator and parameter
            set information to generate candidates. Each tuple is of the form:
            ('name' <str>, 'estimator' <sklearn Estimator>, 'param_set' <dict>)
            For example: ('rf', RandomForestClassifier(), {'max_depth': [5,10]})
        sc (sparkContext): Spark context for spark broadcasting and rdd operations.
        partitions (int or 'auto'): default 'auto'
            Number of partitions to use for parallelization of parameter
            search space. Integer values or None will be used directly for `numSlices`,
            while 'auto' will set `numSlices` to the number required fits.
        n (int): Number of parameter sets to sample from parameter space for each
            estimator.
        cv (int, cross-validation generator or an iterable): Determines the 
            cross-validation splitting strategy.
        scoring (string, callable, list/tuple, dict or None): A single string or a 
            callable to evaluate the predictions on the test set. If None, 
            the estimator's score method is used.
        random_state (int): Random state used throughout to ensure consistent runs.
        verbose (int, bool): Used to indicate level of stdout logging. 
        refit (bool): Refits best estimator at the end of fit method.
        n_jobs (int): Number of jobs to run in parallel. Only used if sc=None.
            ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
            ``-1`` means using all processors. 
        pre_dispatch (int): Controls the number of jobs that get dispatched 
            during parallel execution. Reducing this number can be useful 
            to avoid an explosion of memory consumption when more jobs 
            get dispatched than CPUs can process. Only used if sc=None. 
    """
    def __init__(self, 
                 models,
                 sc=None,
                 partitions="auto",
                 n=5,
                 cv=5,
                 scoring=None,
                 random_state=None,
                 verbose=0, 
                 refit=True, 
                 n_jobs=None, 
                 pre_dispatch='2*n_jobs'):
        self.models = models
        self.sc = sc
        self.partitions = partitions
        self.n = n
        self.cv = cv
        self.scoring = scoring
        self.random_state = random_state
        self.verbose = verbose
        self.refit = refit 
        self.n_jobs = n_jobs 
        self.pre_dispatch = pre_dispatch
        
    def fit(self, X, y=None, groups=None, **fit_params):
        """
        Run fit with all sets of parameters. Parallelize fit operations
        using spark.
        
        Args:
            X (array-like, shape = [n_samples, n_features]): training vector, 
                where n_samples is the number of samples and
                n_features is the number of features
            y (array-like, shape = [n_samples] or [n_samples, n_output]): target 
                relative to X for classification or regression
            groups (array-like, with shape (n_samples,)): group labels for 
                the samples used while splitting the dataset into
                train/test set
            **fit_params (dict of string -> object): parameters passed 
                to the ``fit`` method of the estimator
        """
        _check_estimator(self, verbose=self.verbose)
        models = _validate_models(self.models, self)
        cv = check_cv(
            self.cv, y, classifier=is_classifier(models[0][1]))
        folds = list(cv.split(X,y,groups))
        results = pd.DataFrame()
        
        def _sample_generator(models, n, random_state): 
            rs = None if random_state is None else random_state*(i+1)
            yield _raw_sampler(models, n=n, random_state=random_state)
                
        sample_gen = _sample_generator(
            models, n=self.n, random_state=self.random_state)
        param_sets = list(sample_gen)[0]
        results, model_results = _fit_batch(
            X, y, folds, param_sets, models, self.n, 
            self.scoring, sc=self.sc, n_jobs=self.n_jobs,
            partitions=self.partitions,
            random_state=self.random_state, 
            fit_params=fit_params
            )
        if self.verbose:
            print(model_results)
        
        best_index = np.argmax(results["score"].values)
        self.best_model_index_ = results.iloc[best_index]["model_index"]
        self.best_model_name_ = models[self.best_model_index_][0]
        self.best_params_ = results.iloc[best_index]["param_set"]
        self.best_score_ = results.iloc[best_index]["score"]
        self.worst_score_ = results.iloc[best_index]["score"]
        
        results["rank_test_score"] = np.asarray(
            rankdata(-results["score"].values), 
            dtype=np.int32
            )
        results["mean_test_score"] = results["score"]
        results["params"] = results["param_set"]
        results["model_name"] = (
            results["model_index"]
            .apply(lambda x: models[x][0])
            )
        result_cols = [
            "model_index", "model_name", "params", 
            "rank_test_score", "mean_test_score"
            ]
        self.cv_results_ = results[result_cols].to_dict(orient="list")
        
        if self.refit:
            self.best_estimator_ = _clone(models[self.best_model_index_][1])
            self.best_estimator_.set_params(**self.best_params_)
            self.best_estimator_.fit(X, y)
        
        del self.sc
        return self
    
    def _check_is_fitted(self, method_name):
        if not self.refit:
            raise NotFittedError('This %s instance was initialized '
                                 'with refit=False. %s is '
                                 'available only after refitting on the best '
                                 'parameters. You can refit an estimator '
                                 'manually using the ``best_params_`` '
                                 'attribute'
                                 % (type(self).__name__, method_name))
        else:
            check_is_fitted(self, 'best_estimator_')
    
    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def predict(self, X):
        self._check_is_fitted('predict')
        return self.best_estimator_.predict(X)

    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def predict_proba(self, X):
        self._check_is_fitted('predict_proba')
        return self.best_estimator_.predict_proba(X)

    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def predict_log_proba(self, X):
        self._check_is_fitted('predict_log_proba')
        return self.best_estimator_.predict_log_proba(X)

    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def decision_function(self, X):
        self._check_is_fitted('decision_function')
        return self.best_estimator_.decision_function(X)

    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def transform(self, X):
        self._check_is_fitted('transform')
        return self.best_estimator_.transform(X)

    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def inverse_transform(self, Xt):
        self._check_is_fitted('inverse_transform')
        return self.best_estimator_.inverse_transform(Xt)

    @property
    def classes_(self):
        self._check_is_fitted("classes_")
        return self.best_estimator_.classes_



"""
Functions for generating vectorized spark UDFs
to distribute the prediction of sklearn model predictions
with PySpark DataFrames
"""

import pandas as pd
import numpy as np

class PysparkRequired(ImportError):
    pass

class PyarrowRequired(ImportError):
    pass

_PYSPARK_INSTALLED = None
_PYARROW_INSTALLED = None

def _is_pyspark_installed():
    global _PYSPARK_INSTALLED
    if _PYSPARK_INSTALLED is None:
        try:
            import pyspark
            _PYSPARK_INSTALLED = True
        except ImportError:
            _PYSPARK_INSTALLED = False

    if _PYSPARK_INSTALLED:
        import pyspark
        return True
    else:
        return False

def _is_pyarrow_installed():
    global _PYARROW_INSTALLED
    if _PYARROW_INSTALLED is None:
        try:
            import pyarrow
            _PYARROW_INSTALLED = True
        except ImportError:
            _PYARROW_INSTALLED = False

    if _PYARROW_INSTALLED:
        import pyarrow
        return True
    else:
        return False

def _get_vals(*cols, feature_type="numpy", names=None):
    """ Prep input data for prediction method """
    if feature_type == "numpy":
        vals = np.transpose([a.values for a in cols])
    elif feature_type == "pandas":
        if names is None:
            raise Exception("Must pass names argument for pandas feature type")
        vals = pd.DataFrame(np.transpose([a.values for a in cols]), columns=names)
    elif feature_type == "text":
        vals = cols[0].values
    else:
        raise ValueError("Unknown feature_type: {0}".format(feature_type))
    return vals

def get_prediction_udf(model, method="predict", feature_type="numpy", names=None):
    """
    Build a vectorized PySpark UDF to apply a sklearn model's `predict` or 
    `predict_proba` methods columns in a PySpark DataFrame. Handles
    flexible types of feature data for prediction including 2-D numpy
    arrays ('numpy'), single field text data ('text') and mixed type
    pandas DataFrames ('pandas'). The UDF can then be applied as shown in the 
    example below.

    NOTE: This function requires pyarrow and pyspark with appropriate
    versions for vectorized pandas UDFs and appropriate spark configuration
    to use pyarrow. Ths requires pyarrow>=0.8.0 and pyspark>=2.3.0. 
    Additionally, the spark version must be 2.3 or higher. These requirements 
    are not enforced by the sk-dist package at setup time.

    Args:
        model (sklearn Estimator): sklearn model to distribute
            predictions with PySpark
        method (str): name of prediction method; either 'predict'
            or 'predict_proba'
        feature_type (str): name of feature type; either 'numpy',
            'pandas' or 'text'
        names (array-like): list of ordered column names
            (only necessary for 'pandas' feature_type
    Returns:
        PySpark pandas UDF (pyspark.sql.functions.pandas_udf)    
    Example:
    >>> import pandas as pd
    >>> from sklearn.datasets import load_digits
    >>> from sklearn.linear_model import LogisticRegression
    >>> from pyspark.sql import SparkSession
    >>> spark = (
    >>>     SparkSession
    >>>     .builder
    >>>     .getOrCreate()
    >>>     )
    >>> data = load_digits()
    >>> X = data["data"]
    >>> y = data["target"]
    >>> model = LogisticRegression()
    >>> model.fit(X, y)
    >>> predict = get_prediction_udf(model, method="predict")
    >>> predict_proba = get_prediction_udf(model, method="predict_proba")
    >>> pdf = pd.DataFrame(X)
    >>> sdf = spark.createDataFrame(pdf)
    >>> cols = [F.col(str(c)) for c in sdf.columns]
    >>> prediction_df = (
    >>>     sdf
    >>>     .withColumn("scores", predict_proba(*cols))
    >>>     .withColumn("preds", predict(*cols))
    >>>     .select("preds", "scores")
    >>>     )
    >>> prediction_df.show() 
    ... +-----+--------------------+
    ... |preds|              scores|
    ... +-----+--------------------+
    ... |    0|[0.99988026795692...|
    ... |    1|[4.75035277837040...|
    ... |    2|[2.94811218592164...|
    ... |    3|[1.63438595023762...|
    ... |    4|[1.11339868338047...|
    ... |    5|[1.47300432716012...|
    ... |    6|[1.08560009259480...|
    ... |    7|[3.02428232165044...|
    ... |    8|[7.65445972596079...|
    ... |    9|[3.97610488897298...|
    ... |    0|[0.99918670844137...|
    ... |    1|[2.65336456879078...|
    ... |    2|[1.85886361541580...|
    ... |    3|[2.89824009324990...|
    ... |    4|[2.84813979824305...|
    ... |    5|[2.70090567992820...|
    ... |    6|[1.10907772018062...|
    ... |    7|[3.06455862370095...|
    ... |    8|[2.38739344440480...|
    ... |    9|[8.23628591704589...|
    ... +-----+--------------------+
    ... only showing top 20 rows   
    """
    if not _is_pyspark_installed():
        raise ModuleNotFoundError("Module pyspark not found")
    if not _is_pyarrow_installed():
        raise ModuleNotFoundError("Module pyarrow not found")
    from pyspark.sql import functions as F
    from pyspark.sql.types import (
        DoubleType, StringType, IntegerType, ArrayType
        )

    if method == "predict":
        def predict_func(*cols):
            vals = _get_vals(
                *cols, feature_type=feature_type, 
                names=names
                )
            return pd.Series(model.predict(vals))
        return_type = (
            StringType() 
            if isinstance(model.classes_[0], str) 
            else IntegerType()
            )
        predict = F.pandas_udf(predict_func, returnType=return_type)
    elif method == "predict_proba":  
        def predict_func(*cols):
            vals = _get_vals(
                *cols, feature_type=feature_type, 
                names=names
                )
            return pd.Series(list(model.predict_proba(vals)))
        predict = F.pandas_udf(predict_func, returnType=ArrayType(DoubleType()))
    else:
        raise ValueError("Unknown method: {0}".format(method))
    return predict



"""
Distributed, flexible feature encoder
"""

import ast
import numpy as np

from pandas import DataFrame
from scipy import sparse
from copy import copy
from joblib import Parallel, delayed

from sklearn.pipeline import FeatureUnion
from sklearn.base import TransformerMixin, BaseEstimator

from .base import _parse_partitions 

__all__ = [
    "Encoderizer",
    "EncoderizerExtractor"
]

def _transform_one(transformer, weight, X):
    res = transformer.transform(X)
    # if we have a weight for this transformer, multiply output
    if weight is None:
        return res
    return res * weight

def _fit_one_transformer(transformer, X, y):
    return transformer.fit(X, y)

class Encoderizer(FeatureUnion):
    """ 
    Generic feature encoder with flexible data type input, 
    optional bottom up encoder type inference, and 
    top down `FeatureUnion` configuration acceptance. Also
    includes feature origin method to track down the origin
    transformer of a given feature. Can optionally distribute
    feature union fitting using spark.
    
    Includes `extract` method for slicing off peices
    of the fitted `FeatureUnion` pipeline as a copy of 
    itself with only the requested steps.

    Args:
        transformer_list (iterable or array-like):
            List of transformers, similar to that of FeatureUnion
        transformer_wieghts (dict):
            Multiplicative weights for features per transformer. 
            Keys are transformer names, values the weights.
        n_jobs (int):
            Number of jobs for joblib parallelization
        size (string, 'small', 'medium', or 'large')
            Size of default encoder if using encoder inference
        col_names (iterable or array-like):
            List of column names if using numpy input
        config (dict):
            Configuration of column data types. Keys are column names
            and values are names of default encoder types.
        sc (sparkContext): Spark context for spark broadcasting and rdd operations.
        partitions (int or 'auto'): default 'auto'
            Number of partitions to use for parallelization of parameter
            search space. Integer values or None will be used directly for `numSlices`,
            while 'auto' will set `numSlices` to the number required fits.
    """
    def __init__(self, 
            transformer_list=None, transformer_weights=None,
            n_jobs=1, size="small", config=None,
            col_names=None, sc=None, partitions='auto'):
        self.transformer_list = transformer_list
        self.transformer_weights = transformer_weights
        self.n_jobs = n_jobs
        self.size = size
        self.config = config
        self.col_names = col_names
        self.sc = sc
        self.partitions = partitions
            
    def extract(self, step_names):
        """ 
        Extract copy of fitted self with slice of transformer list
        
        Args:
            step_names (list): White list of transformer
            names to include in extraction slice
            
        Returns:
            fitted Encoderizer object
        """
        encoderizer = copy(self)
        encoderizer.transformer_lengths = [
            encoderizer.transformer_lengths[x] 
            for x in range(len(encoderizer.step_names))
            if encoderizer.step_names[x] in step_names
            ]
        encoderizer.transformer_list = [
            encoderizer.transformer_list[x]
            for x in range(len(encoderizer.step_names))
            if encoderizer.step_names[x] in step_names
            ]        
        return encoderizer
    
    def fit(self, X, y=None):
        """
        Fit all transformers using X

        Args:
            X (iterable or array-like):
                Input data, used to fit transformers.
            y (array-like, shape (n_samples, ...), optional):
               Targets for supervised learning.
        Returns:
            self (Encoderizer object)
                This estimator
        """
        X = self._process_input(X)
        if self.transformer_list is None:
            self.transformer_list = self._infer_transformers(X)
        self.transformer_list = list(self.transformer_list)
        self._validate_transformers()
        if self.sc is None:
            transformers = Parallel(n_jobs=self.n_jobs)(
                delayed(_fit_one_transformer)(trans, X, y)
                for _, trans, _ in self._iter())
        else:
            partitions = _parse_partitions(self.partitions, len(self.transformer_list))
            trans_rdd = self.sc.parallelize(
                list(zip(list(range(len(list(self._iter())))), [x[1] for x in self._iter()])), 
                numSlices=partitions)
            def fot(trans, X, y):
                if y is not None:
                    return _fit_one_transformer(trans, X, y)
                else:
                    return _fit_one_transformer(trans, X, None)
            transformers = trans_rdd.map(lambda x: [x[0], fot(x[1], X, y)]).collect()
            indices = [x[0] for x in transformers]
            trans_objs = [x[1] for x in transformers]
            transformers = list([trans_objs[indices.index(x)] for x in indices])
        self._update_transformer_list(transformers)
        self._feature_indices(X)
        del self.sc
        return self
        
    def transform(self, X):
        """
        Transform X separately by each transformer, concatenate results

        Args:
            X (iterable or array-like):
                Input data to be transformed.
        Returns:
            X_t (array-like or sparse matrix):
                hstack of results of transformers. sum_n_components is the
                sum of n_components (output dimension) over transformers.
        """
        X = self._process_input(X, fit=False)
        Xs = Parallel(n_jobs=self.n_jobs)(
            delayed(_transform_one)(trans, weight, X)
            for name, trans, weight in self._iter())
        if not Xs:
            # All transformers are None
            return np.zeros((X.shape[0], 0))
        if any(sparse.issparse(f) for f in Xs):
            Xs = sparse.hstack(Xs).tocsr()
        else:
            Xs = np.hstack(Xs)
        return Xs

    def fit_transform(self, X, y=None, **fit_params):
        """
        Fit to data, then transform it. Fits transformer to 
        X and y with optional parameters fit_params and returns 
        a transformed version of X.
        
        Args:
            X (numpy array of shape [n_samples, n_features])
                Training set.
            y (numpy array of shape [n_samples]):
                Target values.
   
        Returns:
            X_new (numpy array of shape [n_samples, n_features_new])
                Transformed array.
        """
        if y is None:
            return self.fit(X, **fit_params).transform(X)
        else:
            return self.fit(X, y, **fit_params).transform(X)

    def feature_origin(self, index, mask=None):
        """
        Return step name given a feature index. Optionally 
        supply a custom mask if used in a supplemental feature selection
        step late in the pipeline.

        Args:
            index (int):
                Index of desired feature in transformed
                feature vector.
            mask (numpy ndarray, default None):
                optional boolean mask from feature selector used
                later in the transformation pipeline.
        Returns:
            step_name (string): 
                Name of feature union step of origin
                transformer.
        """
        cumulative = np.cumsum(self.transformer_lengths)
        if mask is not None:
            cumulative = np.array([
                (mask[x - 1]) 
                for x in cumulative])
        return self.step_names[np.argmax(cumulative > index)]
        
    @property
    def step_names(self):
        """ Get ordered list of step names """
        return [x[0] for x in self.transformer_list]
        
    def _process_input(self, X, fit=True):
        """ 
        Converts flexible intput type into pandas DataFrame. Handles
        pandas DataFrame, dictionary, pyspark DataFrame, or 
        numpy ndarray.
        """
        if isinstance(X, DataFrame):
            out = X
        elif isinstance(X, dict):
            try:
                out = DataFrame.from_dict(
                    X, orient="columns")
            except:
                raise ValueError("Cannot parse input")
        elif isinstance(X, np.ndarray) or isinstance(X, list):
            if fit and self.col_names is None:
                raise ValueError(
                    "Must supply col_names with numpy array input")
            elif fit:
                out = DataFrame(X, columns=self.col_names)
            else:
                out = DataFrame(X, columns=self.fields_)
        else:
            from pyspark.sql import DataFrame as SparkDataFrame
            if isinstance(X, SparkDataFrame):
                out = X.toPandas()
            else:
                raise ValueError("Cannot parse input")
        if fit:
            self.fields_ = list(out.columns)
        return out
            
    def _infer_transformers(self, X):
        """ 
        Infer transformer steps based on data types
        and distributions. Use config instead if given.
        """
        from ._defaults import _default_encoders
        if self.config is not None:
            lst = [
                _default_encoders[self.size][v](c)
                for c, v in self.config.items()]
        else:
            lst = [
                self._infer_column(
                    c, X[c], 
                    _default_encoders
                    ) 
                for c in X.columns]
        return [item for sublist in lst if sublist is not None for item in sublist]

    @staticmethod
    def _is_dict(col, col_name):
        """Check if numpy array contains dictionaries, if string attempt to conver to dict"""
        col = col.values
        i=0
        while col[i] is None:  
            i+=1
        col = col[i]
        raise_exception = False
        if isinstance(col, str):
            try:
                ast.literal_eval(col)
                raise_exception = True
            except:
                return False
        if raise_exception:
            raise ValueError("Convert this column to dict before fitting: {0}".format(col_name))
        return isinstance(col, dict)

    @staticmethod
    def _is_list(col, col_name):
        """Check if numpy array contains lists of strings, if string attempt to conver to list"""
        col = col.values
        i=0 
        while col[i] is None:
            i+=1
        col = col[i]
        raise_exception = False
        if isinstance(col, str):
            try:
                ast.literal_eval(col)
                raise_exception = True
            except:
                return False
        if raise_exception:
            raise ValueError("Convert this column to list before fitting: {0}".format(col_name))
        return isinstance(col, list)

    @staticmethod
    def _is_tuple(col, col_name):
        """Check if numpy array contains tuples of strings, if string attempt to conver to tuple"""
        col = col.values
        i=0 
        while col[i] is None:
            i+=1
        col = col[i]
        raise_exception = False
        if isinstance(col, str):
            try:
                ast.literal_eval(col)
                raise_exception = True
            except:
                return False
        if raise_exception:
            raise ValueError("Convert this column to tuple before fitting: {0}".format(col_name))
        return isinstance(col, tuple)

    def _infer_column(self, col_name, X, _default_encoders, thresh=0.10):
        """ Infer encoder type of individual DataFrame column """
        if np.all(X.values == None):
            assert Warning('Column is entirely null: {0}'.format(col_name)) 
            return None       

        is_dict = self._is_dict(X, col_name)
        if is_dict:
            return _default_encoders[self.size]["dict"](col_name)

        is_list = self._is_list(X, col_name)
        if is_list:
            return _default_encoders[self.size]["multihotencoder"](col_name)

        is_tuple = self._is_tuple(X, col_name)
        if is_tuple:
            return _default_encoders[self.size]["multihotencoder"](col_name)
        
        try:
            np.mean(X)
            is_numeric = True
        except:
            is_numeric = False
    
        num_obs = float(len(X))
        pct_unique = len(X.unique()) / num_obs
        is_categorical = pct_unique < thresh
        
        if not is_numeric and not is_categorical:
            return _default_encoders[self.size]["string_vectorizer"](col_name)
        elif is_numeric and not is_categorical:
            return _default_encoders[self.size]["numeric"](col_name)
        else:
            return _default_encoders[self.size]["onehotencoder"](col_name)

    def _feature_indices(self, X):
        """ Save transformed feature vector sizes per transformer """
        projections = [
            transformer.transform(X.head(1))
            for _, transformer, _ 
            in self._iter()]
        
        self.transformer_lengths = [
            len(x[0]) if isinstance(x, list) else x.shape[1]
            for x in projections]

class EncoderizerExtractor(TransformerMixin, BaseEstimator):
    """
    Transformer pass through used for hyperparameter optimization
    in a pipeline
    
    Args:
        encoderizer (fitted Encoderizer instance):
            Encoder from which to slice
        step_names (array-like):
            List of step_names to extract from encoderizer
    """
    def __init__(self, encoderizer, step_names):
        self.encoderizer = encoderizer.extract(step_names)
        
    def fit(self, X, y=None):
        """ Trivial fit method """
        return self

    def transform(self, X):
        """ Extract appropriate transformation steps from Encoderizer """
        return self.encoderizer.transform(X)



"""
Default feature encoding functions and pipelines
for automated feature transformation.
"""

from sklearn.pipeline import Pipeline 
from sklearn.preprocessing import OneHotEncoder
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.feature_extraction import DictVectorizer
from sklearn.feature_selection import VarianceThreshold
from sklearn.preprocessing import StandardScaler
from sklearn.impute import SimpleImputer

from ..preprocessing import (
    ImputeNull, SelectField, FeatureCast, 
    LabelEncoderPipe, HashingVectorizerChunked,
    MultihotEncoder
    )

def tokenizer(x):
    """ Trivial tokenizer """
    return x

dict_encoder = lambda c: [(
                "{0}_dict_encoder".format(c),
                Pipeline(steps=[
                    ("var", SelectField(cols=[c], single_dimension=True)),
                    ("fillna", ImputeNull({})),
                    ("vec", DictVectorizer())])  
                )]

onehot_encoder = lambda c: [(
                "{0}_onehot".format(c), 
                 Pipeline(steps=[
                     ("var", SelectField(cols=[c], single_dimension=True)),
                     ("cast", FeatureCast(cast_type=str)),
                     ("fillna", ImputeNull("")),
                     ("vec", CountVectorizer(
                         tokenizer=tokenizer, 
                         binary=True,
                         decode_error="ignore"))])
                )]

multihot_encoder = lambda c: [(
                "{0}_multihot".format(c), 
                Pipeline(steps=[
                   ("var", SelectField(cols=[c], single_dimension=True)),
                   ("fillna", ImputeNull([])),
                   ("vec", MultihotEncoder())])
                )]

numeric_encoder = lambda c: [(
                "{0}_scaler".format(c),
                Pipeline(steps=[
                    ("var", SelectField(cols=[c])),
                    ("imputer", SimpleImputer(strategy="median")), 
                    ("scaler", StandardScaler(copy=False))])
                )]

_default_encoders = {
    "small": {
        "string_vectorizer": lambda c: [(
                "{0}_word_vec".format(c), 
                 Pipeline(steps=[
                     ("var", SelectField(cols=[c], single_dimension=True)),
                     ("fillna", ImputeNull(" ")), 
                     ("vec", HashingVectorizerChunked(
                         ngram_range=(1,2), 
                         analyzer="word", 
                         decode_error="ignore")),
                     ("var_thresh", VarianceThreshold())])
                )],
        "onehotencoder": onehot_encoder,
        "multihotencoder": multihot_encoder,
        "numeric": numeric_encoder,
        "dict": dict_encoder
    },
    "medium": {
        "string_vectorizer": lambda c: [(
                    "{0}_word_vec".format(c), 
                     Pipeline(steps=[
                         ("var", SelectField(cols=[c], single_dimension=True)),
                         ("fillna", ImputeNull(" ")), 
                         ("vec", HashingVectorizerChunked(
                             ngram_range=(1,3), 
                             analyzer="word", 
                             decode_error="ignore")),
                         ("var_thresh", VarianceThreshold())])
                    ),
                    (
                    "{0}_char_vec".format(c), 
                     Pipeline(steps=[
                         ("var", SelectField(cols=[c], single_dimension=True)),
                         ("fillna", ImputeNull(" ")), 
                         ("vec", HashingVectorizerChunked(
                             ngram_range=(3,4), 
                             analyzer="char_wb", 
                             decode_error="ignore")),
                         ("var_thresh", VarianceThreshold())])
                    )
                ],
        "onehotencoder": onehot_encoder,
        "multihotencoder": multihot_encoder,
        "numeric": numeric_encoder,
        "dict": dict_encoder
    },
    "large": {
        "string_vectorizer": lambda c: [(
                    "{0}_word_vec".format(c), 
                     Pipeline(steps=[
                         ("var", SelectField(cols=[c], single_dimension=True)),
                         ("fillna", ImputeNull(" ")), 
                         ("vec", HashingVectorizerChunked(
                             ngram_range=(1,3), 
                             analyzer="word", 
                             decode_error="ignore")),
                         ("var_thresh", VarianceThreshold())])
                    ),
                    (
                    "{0}_char_vec".format(c), 
                     Pipeline(steps=[
                         ("var", SelectField(cols=[c], single_dimension=True)),
                         ("fillna", ImputeNull(" ")), 
                         ("vec", HashingVectorizerChunked(
                             ngram_range=(2,5), 
                             analyzer="char_wb", 
                             decode_error="ignore")),
                         ("var_thresh", VarianceThreshold())])
                    )
                ],
        "onehotencoder": onehot_encoder,
        "multihotencoder": multihot_encoder,
        "numeric": numeric_encoder,
        "dict": dict_encoder
    }
}



"""
Base functions for distributed meta-estimators
"""

import copy

def _clone(estimator, safe=True):
    """
    Constructs a new estimator with the same parameters.
    Handles sparkContext which shouldn't be copied.
    """
    found_sc = False
    if hasattr(estimator, "sc"):
        found_sc = True
    estimator_type = type(estimator)
    # XXX: not handling dictionaries
    if estimator_type in (list, tuple, set, frozenset):
        return estimator_type([_clone(e, safe=safe) for e in estimator])
    elif not hasattr(estimator, 'get_params') or isinstance(estimator, type):
        if not safe:
            return copy.deepcopy(estimator)
        else:
            raise TypeError("Cannot clone object '%s' (type %s): "
                            "it does not seem to be a scikit-learn estimator "
                            "as it does not implement a 'get_params' methods."
                            % (repr(estimator), type(estimator)))
    klass = estimator.__class__
    new_object_params = estimator.get_params(deep=False)
    for name, param in new_object_params.items():
        if name != "sc":
            new_object_params[name] = _clone(param, safe=False)
    new_object = klass(**new_object_params)
    params_set = new_object.get_params(deep=False)

    # quick sanity check of the parameters of the clone
    for name in new_object_params:
        param1 = new_object_params[name]
        param2 = params_set[name]
        if param1 is not param2:
            raise RuntimeError('Cannot clone object %s, as the constructor '
                               'either does not set or modifies parameter %s' %
                               (estimator, name))

    if found_sc:
        new_object.sc = estimator.sc
    return new_object

def _parse_partitions(partitions, auto_n):
    """ Handles the partitions input for spark parallelization """
    if partitions is None:
        partitions = None
    elif partitions == 'auto':
        partitions = auto_n
    else:
        try:
            partitions = int(partitions)
        except:
            partitions = None
    return partitions

def _get_value(obj):
    """ 
    Determines if input object is broadcast variable.
    If so return value attribute, else return object.
    """
    return obj.value if hasattr(obj, "value") else obj



"""
Test encoder classes
"""

try:
    import pandas as pd
    import numpy as np
    from skdist.distribute import encoder
    from skdist.distribute.encoder import Encoderizer, EncoderizerExtractor
    _import_error = None
except Exception as e:
    _import_error = e

def test_encoder():
    assert _import_error == None

def test_simple():
    # simple encoderizer example
    X = [
        "text encoding is fun", 
        "this is a text encoding example",
        "hopefully it works and we'll have passed the test"
        ]
    encoderizer = encoder.Encoderizer(col_names=["text"])
    X_t = encoderizer.fit_transform(X)
    assert X_t.shape == (3,31)

def test_pandas():
    # pandas encoderizer example
    X = [
        "text encoding is fun",
        "this is a text encoding example",
        "hopefully it works and we'll have passed the test"
        ]
    df = pd.DataFrame({
        "text": X*4, 
        "constant_int": [np.nan,2,2]*4, 
        "constant_str": ["test","test","test"]*4, 
        "constant_null": [None,"test",np.nan]*4, 
        "numbers": [0.123, None, 0.535]*4,
        "dicts": [{"a": 4}, None, {"b": 1}]*4
        })
    encoderizer = encoder.Encoderizer(size="medium")
    X_t = encoderizer.fit_transform(df)
    assert X_t.shape == (12, 180)

def test_dict_list():
    # pandas encoderizer example
    X = [
        "[text encoding is fun]",
        "this is a text encoding {'pizza': 'example'}",
        "hopefully it works and [we'll have passed] the test"
        ]
    df = pd.DataFrame({
        "text": X*4,
        "constant_str": ["test","test","test"]*4,
        "dicts": [{"a": 4}, None, {"b": 1}]*4,
        "lists": [["this", "is", "text"], ["more", "text"], ["text"]]*4
        })
    encoderizer = encoder.Encoderizer(size="medium")
    X_t = encoderizer.fit_transform(df)
    assert X_t.shape == (12, 210)

def test_dict_list_tuple():
    # pandas encoderizer example
    X = [
        "[text encoding is fun]",
        "this is a text encoding {'pizza': 'example'}",
        "hopefully it works and [we'll have passed] the test"
        ]
    df = pd.DataFrame({
	    "text_col": X*4,
	    "categorical_str_col": ["control", "treatment", "control"]*4,
	    "categorical_int_col": [0, 1, 2]*4,
	    "numeric_col": [5, 22, 69]*4,
	    "dict_col": [{"a": 4}, {"b": 1}, {"c": 3}]*4,
	    "list_col": [[1, 2], [1, 3], [2,]]*4,
	    "tuple_col": [(1, 2), (1, 3), (2,)]*4
	    })
    encoderizer = encoder.Encoderizer(size="medium")
    X_t = encoderizer.fit_transform(df)
    assert X_t.shape == (12, 244)
    


"""
Run setup
"""

from setuptools import setup, find_packages
from skdist import __version__

DISTNAME = "sk-dist"
VERSION = __version__
DESCRIPTION = "Distributed scikit-learn meta-estimators with PySpark"
with open("README.rst") as f:
    LONG_DESCRIPTION = f.read()
CLASSIFIERS = [
    "Intended Audience :: Developers",
    "Programming Language :: Python",
    "Topic :: Scientific/Engineering"
    ]
AUTHOR = "Ibotta Inc."
AUTHOR_EMAIL = "machine_learning@ibotta.com"
LICENSE = "Apache 2.0"
DOWNLOAD_URL = "https://pypi.org/project/sk-dist/#files"
PROJECT_URLS = {
    "Source Code": "https://github.com/Ibotta/sk-dist"
    }
MIN_PYTHON_VERSION = "3.5"
MIN_PANDAS_VERSION = "0.19.0"
MIN_NUMPY_VERSION = "0.17.0"
MIN_SCIPY_VERSION = "1.3.1"
MIN_SKLEARN_VERSION = "0.21.3"
MIN_JOBLIB_VERSION = "0.11"
MIN_PYARROW_VERSION = "0.8.0"
MIN_PYSPARK_VERSION = "2.4.4"
MIN_PYTESTSPARK_VERSION = "0.4.5"

install_requires = [
    "pandas>={0}".format(MIN_PANDAS_VERSION),
    "numpy>={0}".format(MIN_NUMPY_VERSION),
    "scipy>={0}".format(MIN_SCIPY_VERSION),
    "scikit-learn>={0}".format(MIN_SKLEARN_VERSION),
    "joblib>={0}".format(MIN_JOBLIB_VERSION)
]

tests_require = [
    "pyarrow>={0}".format(MIN_PYARROW_VERSION),
    "pyspark>={0}".format(MIN_PYSPARK_VERSION),
    "pytest-spark>={0}".format(MIN_PYTESTSPARK_VERSION)
]

def parse_description(description):
    """
    Strip figures and alt text from description
    """
    return "\n".join(
        [
        a for a in description.split("\n")
        if ("figure::" not in a) and (":alt:" not in a)
        ])

setup(name=DISTNAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=parse_description(LONG_DESCRIPTION),
      classifiers=CLASSIFIERS,
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      license=LICENSE,
      download_url=DOWNLOAD_URL,
      project_urls=PROJECT_URLS,
      packages=find_packages(),
      python_requires=">={0}".format(MIN_PYTHON_VERSION),
      install_requires=install_requires,
      tests_require=tests_require,
      extras_require=dict(tests=tests_require)
      )


"""
Various postprocessing classes implemented as scikit-learn
transformers compatible with scikit-learn pipelines.
"""

import numpy as np

from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.preprocessing import LabelEncoder
from sklearn.utils.validation import check_is_fitted
from sklearn.utils import Bunch

__all__ = [
    "SimpleVoter"
    ]

class SimpleVoter(BaseEstimator, ClassifierMixin):
    """
    Simple voting classifier which takes fitted estimators as
    input and applies voting ensembling at predict time. Similar to
    `sklearn.enseble.VotingClassifier` except the VotingClassifier
    will actually fit the child estimators, where the SimpleVoter
    will take fitted estimators at instantiation and apply similar
    prediction functions as predict time. 
    
    The VotingClassifier implements a dummy fit method that does nothing
    unless the class attributes `estimators` or `classes` have been 
    manually updated.
    
    The value add over `sklearn.enseble.VotingClassifier` is that
    SimpleVoter can solely run the predict, and let the fit live elsewhere.
    
    Args:
        estimators (list of tuples): list of fitted (name, estimator)
            tuples for voting
        classes (array-like): list of class names from one
            of the underlying estimators
        voting (str): voting method ('hard' or 'soft')
        weights (array-like): array of weights corresponding
            to each estimator
    
    """
    def __init__(self, estimators, classes, voting='hard', weights=None):
        self.estimators = estimators
        self.classes = classes
        self.voting = voting
        self.weights = weights
        self._assemble_attributes()
        
    @property
    def named_estimators(self):
        """ Bunches the estimators by name """
        return Bunch(**dict(self.estimators))
        
    @property
    def _weights_not_none(self):
        """Get the weights of not `None` estimators"""
        if self.weights is None:
            return None
        return [w for est, w in zip(self.estimators, self.weights)
                if est[1] not in (None, 'drop')]
        
    def fit(self, X, y=None):
        """ Trivial fit method; re-assembles attributes """
        self._assemble_attributes()
        return self

    def predict(self, X):
        """ Compute predictions for samples in X """
        check_is_fitted(self, 'estimators_')
        if self.voting == 'soft':
            maj = np.argmax(self.predict_proba(X), axis=1)
        else:  # 'hard' voting
            predictions = self._predict(X)
            maj = np.apply_along_axis(
                lambda x: np.argmax(
                    np.bincount(x, weights=self._weights_not_none)),
                axis=1, arr=predictions)
        maj = self.le_.inverse_transform(maj)
        return maj
        
    @property
    def predict_proba(self):
        """ Compute probabilities of possible outcomes for samples in X """
        return self._predict_proba
        
    def _predict(self, X):
        "" "Collect results from clf.predict calls """
        return np.asarray([clf.predict(X) for clf in self.estimators_]).T
        
    def _predict_proba(self, X):
        """ Predict class probabilities for X in 'soft' voting """
        if self.voting == 'hard':
            raise AttributeError("predict_proba is not available when"
                                 " voting=%r" % self.voting)
        check_is_fitted(self, 'estimators_')
        avg = np.average(self._collect_probas(X), axis=0,
                         weights=self._weights_not_none)
        return avg
        
    def _collect_probas(self, X):
        """Collect results from clf.predict calls. """
        return np.asarray([clf.predict_proba(X) for clf in self.estimators_])
        
    def _assemble_attributes(self):
        """ Assemble fitted class attributes """
        names, clfs = zip(*self.estimators)
        self.estimators_ = clfs
        self.classes_ = self.classes
        self.le_ = LabelEncoder()
        self.le_.classes_ = self.classes



"""
Test predict imports
"""

try:
    from skdist.distribute import predict
    from skdist.distribute.predict import get_prediction_udf 
    _import_error = None
except Exception as e:
    _import_error = e

def test_predict():
    assert _import_error == None
